from io import BytesIO
import math
import fastavro
from fastavro.io.binary_decoder import BinaryDecoder
from fastavro.read import _read as _reader, HEADER_SCHEMA, SchemaResolutionError
from fastavro.write import _write as _writer, Writer

import pytest

import copy
import datetime
import sys
import traceback
import zipfile
from collections import OrderedDict
from os.path import join, abspath, dirname, basename
from glob import iglob

data_dir = join(abspath(dirname(__file__)), "avro-files")

try:
    import snappy  # NOQA

    has_snappy = True
except ImportError:
    has_snappy = False

NO_DATA = {
    "class org.apache.avro.tool.TestDataFileTools.zerojsonvalues.avro",
    "testDataFileMeta.avro",
}


def roundtrip(schema, records, *, writer_kwargs={}, **reader_kwargs):
    new_file = BytesIO()
    fastavro.writer(new_file, schema, records, **writer_kwargs)
    new_file.seek(0)

    reader = fastavro.reader(new_file, **reader_kwargs)

    new_records = list(reader)
    return new_records


class NoSeekBytesIO:
    """Shim around BytesIO which blocks access to everything but read.
    Used to ensure seek and tell API isn't being depended on."""

    def __init__(self, *args):
        self.underlying = BytesIO(*args)

    def read(self, n):
        return self.underlying.read(n)

    def tell(self):
        raise AssertionError("fastavro reader should not depend on tell")

    def seek(self, *args):
        raise AssertionError("fastavro reader should not depend on seek")


def _test_files():
    for filename in iglob(join(data_dir, "*.avro")):
        if (not has_snappy) and ("snappy" in filename):
            continue
        yield filename


def remove_legacy_fields(schema):
    if "__fastavro_parsed" in schema:
        schema.pop("__fastavro_parsed")
    if "__named_schemas" in schema:
        schema.pop("__named_schemas")
    return schema


@pytest.mark.parametrize("filename", _test_files())
def test_file(filename):
    with open(filename, "rb") as fo:
        reader = fastavro.reader(fo)
        assert hasattr(reader, "writer_schema"), "no schema on file"

        if basename(filename) in NO_DATA:
            return

        records = list(reader)
        assert len(records) > 0, "no records found"

    new_file = BytesIO()
    fastavro.writer(new_file, reader.writer_schema, records, reader.codec)
    new_file_bytes = new_file.getvalue()

    new_file = NoSeekBytesIO(new_file_bytes)
    new_reader = fastavro.reader(new_file)
    assert hasattr(new_reader, "writer_schema"), "schema wasn't written"
    assert new_reader.writer_schema == remove_legacy_fields(
        copy.deepcopy(reader.writer_schema)
    )
    assert new_reader.codec == reader.codec
    new_records = list(new_reader)

    assert new_records == records

    # Test schema migration with the same schema
    new_file = NoSeekBytesIO(new_file_bytes)
    schema_migration_reader = fastavro.reader(new_file, reader.writer_schema)
    assert schema_migration_reader.reader_schema == reader.writer_schema
    new_records = list(schema_migration_reader)

    assert new_records == records


def test_not_avro():
    with pytest.raises(ValueError):
        with open(__file__, "rb") as fo:
            fastavro.reader(fo)


def test_parse_schema_rejects_undleclared_name():
    try:
        fastavro.parse_schema(
            {
                "type": "record",
                "name": "test_parse_schema_rejects_undleclared_name",
                "fields": [
                    {
                        "name": "left",
                        "type": "Thinger",
                    }
                ],
            }
        )
        assert False, "Never raised"
    except fastavro.schema.UnknownType as e:
        assert "Thinger" == e.name


def test_parse_schema_rejects_unordered_references():
    try:
        fastavro.parse_schema(
            {
                "type": "record",
                "name": "test_parse_schema_rejects_unordered_references",
                "fields": [
                    {"name": "left", "type": "Thinger"},
                    {
                        "name": "right",
                        "type": {
                            "type": "record",
                            "name": "Thinger",
                            "fields": [{"name": "the_thing", "type": "string"}],
                        },
                    },
                ],
            }
        )
        assert False, "Never raised"
    except fastavro.schema.UnknownType as e:
        assert "Thinger" == e.name


def test_parse_schema_accepts_nested_namespaces():
    parsed_schema = fastavro.parse_schema(
        {
            "namespace": "com.example",
            "name": "Outer",
            "type": "record",
            "fields": [
                {
                    "name": "a",
                    "type": {
                        "type": "record",
                        "name": "Inner",
                        "fields": [{"name": "the_thing", "type": "string"}],
                    },
                },
                {
                    "name": "b",
                    # This should resolve to com.example.Inner because of the
                    # `namespace` of the enclosing record.
                    "type": "Inner",
                },
                {"name": "b", "type": "com.example.Inner"},
            ],
        }
    )
    assert "com.example.Inner" == parsed_schema["fields"][0]["type"]["name"]
    assert "com.example.Inner" == parsed_schema["fields"][1]["type"]


def test_parse_schema_resolves_references_from_unions():
    parsed_schema = fastavro.parse_schema(
        {
            "namespace": "com.other",
            "name": "Outer",
            "type": "record",
            "fields": [
                {
                    "name": "a",
                    "type": [
                        "null",
                        {
                            "type": "record",
                            "name": "Inner",
                            "fields": [{"name": "the_thing", "type": "string"}],
                        },
                    ],
                },
                {
                    "name": "b",
                    # This should resolve to com.example.Inner because of the
                    # `namespace` of the enclosing record.
                    "type": ["null", "Inner"],
                },
            ],
        }
    )
    assert "com.other.Inner" == parsed_schema["fields"][1]["type"][1]


def test_parse_schema_accepts_nested_records_from_arrays():
    parsed_schema = fastavro.parse_schema(
        {
            "fields": [
                {
                    "type": {
                        "items": {
                            "type": "record",
                            "fields": [{"type": "string", "name": "text"}],
                            "name": "Nested",
                        },
                        "type": "array",
                    },
                    "name": "multiple",
                },
                {"type": {"type": "array", "items": "Nested"}, "name": "single"},
            ],
            "type": "record",
            "name": "test_parse_schema_accepts_nested_records_from_arrays",
        }
    )
    assert "Nested" == parsed_schema["fields"][1]["type"]["items"]


def test_compose_schemas():
    schema_path = join(data_dir, "Parent.avsc")
    fastavro.schema.load_schema(schema_path)


def test_reading_after_writing_with_load_schema():
    schema_path = join(data_dir, "Parent.avsc")
    schema = fastavro.schema.load_schema(schema_path)

    records = [{"child": {}, "child1": {}}]

    new_file = BytesIO()
    fastavro.writer(new_file, schema, records)
    new_file.seek(0)

    reader = fastavro.reader(new_file)
    new_records = list(reader)
    assert new_records == records


def test_missing_schema():
    schema_path = join(data_dir, "ParentMissingChild.avsc")
    with pytest.raises(fastavro.schema.UnknownType):
        fastavro.schema.load_schema(schema_path)


def test_default_values():
    schema = {
        "type": "record",
        "name": "test_default_values",
        "fields": [
            {"name": "default_field", "type": "string", "default": "default_value"}
        ],
    }
    records = [{}]

    new_records = roundtrip(schema, records)
    assert new_records == [{"default_field": "default_value"}]


def test_nullable_values():
    schema = {
        "type": "record",
        "name": "test_nullable_values",
        "fields": [
            {"name": "nullable_field", "type": ["string", "null"]},
            {"name": "field", "type": "string"},
        ],
    }
    records = [{"field": "val"}, {"field": "val", "nullable_field": "no_null"}]

    new_records = roundtrip(schema, records)
    assert new_records == [
        {"nullable_field": None, "field": "val"},
        {"nullable_field": "no_null", "field": "val"},
    ]


def test_metadata():
    schema = {"type": "record", "name": "test_metadata", "fields": []}

    new_file = BytesIO()
    records = [{}]
    metadata = {"key": "value"}
    fastavro.writer(new_file, schema, records, metadata=metadata)
    new_file.seek(0)
    new_reader = fastavro.reader(new_file)
    assert new_reader.metadata["key"] == metadata["key"]


def test_write_union_shortcut():
    schema = {
        "type": "record",
        "name": "A",
        "fields": [
            {
                "name": "a",
                "type": [
                    {
                        "type": "record",
                        "name": "B",
                        "fields": [{"name": "b", "type": "string"}],
                    },
                    {
                        "type": "record",
                        "name": "C",
                        "fields": [{"name": "c", "type": "string"}],
                    },
                ],
            }
        ],
    }

    records = [{"a": ("B", {"b": "test"})}]

    assert [{"a": {"b": "test"}}] == roundtrip(schema, records)


def test_repo_caching_issue():
    schema = {
        "type": "record",
        "name": "B",
        "fields": [
            {
                "name": "b",
                "type": {
                    "type": "record",
                    "name": "C",
                    "fields": [{"name": "c", "type": "string"}],
                },
            }
        ],
    }

    records = [{"b": {"c": "test"}}]

    assert records == roundtrip(schema, records)

    other_schema = {
        "name": "A",
        "type": "record",
        "fields": [
            {
                "name": "a",
                "type": {
                    "type": "record",
                    "name": "B",
                    "fields": [
                        {
                            "name": "b",
                            "type": {
                                "type": "record",
                                "name": "C",
                                "fields": [{"name": "c", "type": "int"}],
                            },
                        }
                    ],
                },
            },
            {"name": "aa", "type": "B"},
        ],
    }

    records = [{"a": {"b": {"c": 1}}, "aa": {"b": {"c": 2}}}]

    assert records == roundtrip(other_schema, records)


def test_schema_migration_remove_field():
    schema = {
        "type": "record",
        "name": "test_schema_migration_remove_field",
        "fields": [
            {
                "name": "test",
                "type": "string",
            }
        ],
    }

    new_schema = {
        "type": "record",
        "name": "test_schema_migration_remove_field",
        "fields": [],
    }

    new_file = BytesIO()
    records = [{"test": "test"}]
    fastavro.writer(new_file, schema, records)
    new_file.seek(0)
    new_reader = fastavro.reader(new_file, new_schema)
    new_records = list(new_reader)
    assert new_records == [{}]


def test_schema_migration_add_default_field():
    schema = {
        "type": "record",
        "name": "test_schema_migration_add_default_field",
        "fields": [],
    }

    new_schema = {
        "type": "record",
        "name": "test_schema_migration_add_default_field",
        "fields": [
            {
                "name": "test",
                "type": "string",
                "default": "default",
            }
        ],
    }

    new_file = BytesIO()
    records = [{}]
    fastavro.writer(new_file, schema, records)
    new_file.seek(0)
    new_reader = fastavro.reader(new_file, new_schema)
    new_records = list(new_reader)
    assert new_records == [{"test": "default"}]


def test_schema_migration_type_promotion():
    schema = {
        "type": "record",
        "name": "test_schema_migration_type_promotion",
        "fields": [
            {
                "name": "test",
                "type": ["string", "int"],
            }
        ],
    }

    new_schema = {
        "type": "record",
        "name": "test_schema_migration_type_promotion",
        "fields": [
            {
                "name": "test",
                "type": ["float", "string"],
            }
        ],
    }

    new_file = BytesIO()
    records = [{"test": 1}]
    fastavro.writer(new_file, schema, records)
    new_file.seek(0)
    new_reader = fastavro.reader(new_file, new_schema)
    new_records = list(new_reader)
    assert new_records == records


def test_schema_migration_maps_with_union_promotion():
    schema = {
        "type": "record",
        "name": "test_schema_migration_maps_with_union_promotion",
        "fields": [
            {
                "name": "test",
                "type": {"type": "map", "values": ["string", "int"]},
            }
        ],
    }

    new_schema = {
        "type": "record",
        "name": "test_schema_migration_maps_with_union_promotion",
        "fields": [
            {
                "name": "test",
                "type": {"type": "map", "values": ["string", "long"]},
            }
        ],
    }

    new_file = BytesIO()
    records = [{"test": {"foo": 1}}]
    fastavro.writer(new_file, schema, records)
    new_file.seek(0)
    new_reader = fastavro.reader(new_file, new_schema)
    new_records = list(new_reader)
    assert new_records == records


def test_schema_migration_array_with_union_promotion():
    schema = {
        "type": "record",
        "name": "test_schema_migration_array_with_union_promotion",
        "fields": [
            {
                "name": "test",
                "type": {"type": "array", "items": ["boolean", "long"]},
            }
        ],
    }

    new_schema = {
        "type": "record",
        "name": "test_schema_migration_array_with_union_promotion",
        "fields": [
            {
                "name": "test",
                "type": {"type": "array", "items": ["string", "float"]},
            }
        ],
    }

    new_file = BytesIO()
    records = [{"test": [1, 2, 3]}]
    fastavro.writer(new_file, schema, records)
    new_file.seek(0)
    new_reader = fastavro.reader(new_file, new_schema)
    new_records = list(new_reader)
    assert new_records == records


def test_schema_migration_writer_union():
    schema = {
        "type": "record",
        "name": "test_schema_migration_writer_union",
        "fields": [{"name": "test", "type": ["string", "int"]}],
    }

    new_schema = {
        "type": "record",
        "name": "test_schema_migration_writer_union",
        "fields": [{"name": "test", "type": "int"}],
    }

    new_file = BytesIO()
    records = [{"test": 1}]
    fastavro.writer(new_file, schema, records)
    new_file.seek(0)
    new_reader = fastavro.reader(new_file, new_schema)
    new_records = list(new_reader)
    assert new_records == records


def test_schema_migration_reader_union():
    schema = {
        "type": "record",
        "name": "test_schema_migration_reader_union",
        "fields": [{"name": "test", "type": "int"}],
    }

    new_schema = {
        "type": "record",
        "name": "test_schema_migration_reader_union",
        "fields": [{"name": "test", "type": ["string", "int"]}],
    }

    new_file = BytesIO()
    records = [{"test": 1}]
    fastavro.writer(new_file, schema, records)
    new_file.seek(0)
    new_reader = fastavro.reader(new_file, new_schema)
    new_records = list(new_reader)
    assert new_records == records


def test_schema_migration_union_failure():
    schema = {
        "type": "record",
        "name": "test_schema_migration_union_failure",
        "fields": [{"name": "test", "type": "boolean"}],
    }

    new_schema = {
        "type": "record",
        "name": "test_schema_migration_union_failure",
        "fields": [{"name": "test", "type": ["string", "int"]}],
    }

    new_file = BytesIO()
    records = [{"test": True}]
    fastavro.writer(new_file, schema, records)
    new_file.seek(0)
    new_reader = fastavro.reader(new_file, new_schema)

    with pytest.raises(fastavro.read.SchemaResolutionError):
        list(new_reader)


def test_schema_migration_array_failure():
    schema = {
        "type": "record",
        "name": "test_schema_migration_array_failure",
        "fields": [
            {
                "name": "test",
                "type": {"type": "array", "items": ["string", "int"]},
            }
        ],
    }

    new_schema = {
        "type": "record",
        "name": "test_schema_migration_array_failure",
        "fields": [
            {
                "name": "test",
                "type": {"type": "array", "items": ["string", "boolean"]},
            }
        ],
    }

    new_file = BytesIO()
    records = [{"test": [1, 2, 3]}]
    fastavro.writer(new_file, schema, records)
    new_file.seek(0)
    new_reader = fastavro.reader(new_file, new_schema)

    with pytest.raises(fastavro.read.SchemaResolutionError):
        list(new_reader)


def test_schema_migration_maps_failure():
    schema = {
        "type": "record",
        "name": "test_schema_migration_maps_failure",
        "fields": [
            {
                "name": "test",
                "type": {"type": "map", "values": "string"},
            }
        ],
    }

    new_schema = {
        "type": "record",
        "name": "test_schema_migration_maps_failure",
        "fields": [
            {
                "name": "test",
                "type": {"type": "map", "values": "long"},
            }
        ],
    }

    new_file = BytesIO()
    records = [{"test": {"foo": "a"}}]
    fastavro.writer(new_file, schema, records)
    new_file.seek(0)
    new_reader = fastavro.reader(new_file, new_schema)
    with pytest.raises(fastavro.read.SchemaResolutionError):
        list(new_reader)


def test_schema_migration_schema_mismatch():
    schema = {
        "type": "record",
        "name": "test_schema_migration_schema_mismatch",
        "fields": [
            {
                "name": "test",
                "type": "string",
            }
        ],
    }

    new_schema = {
        "type": "enum",
        "name": "test",
        "symbols": ["FOO", "BAR"],
    }

    new_file = BytesIO()
    records = [{"test": "test"}]
    fastavro.writer(new_file, schema, records)
    new_file.seek(0)
    new_reader = fastavro.reader(new_file, new_schema)
    with pytest.raises(fastavro.read.SchemaResolutionError):
        list(new_reader)


def test_empty():
    io = BytesIO()
    with pytest.raises(ValueError, match="cannot read header - is it an avro file?"):
        fastavro.reader(io)


def test_no_default():
    io = BytesIO()
    schema = {
        "type": "record",
        "name": "test",
        "fields": [{"type": "boolean", "name": "a"}],
    }
    with pytest.raises(ValueError, match="no value and no default"):
        fastavro.writer(io, schema, [{}])


def test_is_avro_str():
    for path in iglob(f"{data_dir}/*.avro"):
        assert fastavro.is_avro(path)
    assert not fastavro.is_avro(__file__)


def test_is_avro_fo():
    for path in iglob(f"{data_dir}/*.avro"):
        with open(path, "rb") as fp:
            assert fastavro.is_avro(fp)
    with open(__file__, "rb") as fp:
        assert not fastavro.is_avro(fp)


def test_write_long_union_type():
    schema = {
        "name": "test_name",
        "namespace": "test",
        "type": "record",
        "fields": [
            {"name": "time", "type": ["null", "long"]},
        ],
    }

    records = [
        {"time": 809066167221092352},
    ]

    assert records == roundtrip(schema, records)


def test_cython_python():
    # Since Cython and Python implement the same behavior, it is possible for
    # build errors or coding errors to accidentally result in using the wrong
    # one. This is bad, because the pure Python version is faster in Pypy,
    # while the Cython version is faster in CPython. This test verifies the
    # correct reader and writer implementations are used.
    if hasattr(sys, "pypy_version_info"):
        # Pypy should not use Cython.
        assert not hasattr(_reader, "CYTHON_MODULE")
        assert not hasattr(_writer, "CYTHON_MODULE")
    else:
        # CPython should use Cython.
        assert getattr(_reader, "CYTHON_MODULE")
        assert getattr(_writer, "CYTHON_MODULE")


def test_writer_class_flush_end(tmpdir):
    """
    Create an Avro file using the Writer class. Verify that data accumulates in
    memory and is written when flush() is called.
    """
    schema = {
        "type": "record",
        "name": "Test",
        "namespace": "test",
        "fields": [
            {"name": "field1", "type": {"type": "string"}},
            {"name": "field2", "type": {"type": "int"}},
        ],
    }
    records = [{"field1": "test1", "field2": -1}, {"field1": "test2", "field2": 5}]

    temp_path = tmpdir.join("test_writer_class.avro")
    with temp_path.open("wb") as fo:
        w = Writer(fo, schema, codec="deflate")

        # Creating the Writer adds the Avro file header. Get file size with
        # header only.
        size_with_header_only = fo.tell()
        for i, record in enumerate(records):
            assert w.block_count == i
            w.write(record)

            # Verify records are being stored *in memory*:
            # 1. Block count increases
            # 2. File size does not increase
            assert w.block_count == i + 1
            assert fo.tell() == size_with_header_only

        # Flushing the file writes the data. File size should increase now.
        w.flush()
        assert fo.tell() > size_with_header_only

    # Read the records to verify they were written correctly.
    new_reader = fastavro.reader(temp_path.open("rb"))
    new_records = list(new_reader)
    assert new_records == records


def test_writer_class_sync_interval_automatic_flush(tmpdir):
    """
    Create an Avro file using the Writer class with sync_interval set to 0.
    Verify that data does not accumulate in memory but is automatically flushed
    to the file object as each record is added.
    """
    schema = {
        "type": "record",
        "name": "Test",
        "namespace": "test",
        "fields": [
            {"name": "field1", "type": {"type": "string"}},
            {"name": "field2", "type": {"type": "int"}},
        ],
    }
    records = [{"field1": "test1", "field2": -1}, {"field1": "test2", "field2": 5}]

    temp_path = tmpdir.join("test_writer_class.avro")
    with temp_path.open("wb") as fo:
        w = Writer(fo, schema, codec="deflate", sync_interval=0)

        # Creating the Writer adds the Avro file header. Get file size with
        # header only.
        file_size_history = [fo.tell()]
        for i, record in enumerate(records):
            assert w.block_count == 0
            w.write(record)

            # Verify records are being stored *in memory*:
            # 1. Block count increases
            # 2. File size does not increase
            assert w.block_count == 0
            file_size_history.append(fo.tell())
            assert file_size_history[-1] > file_size_history[-2]

        # Flushing the file writes the data. File size should increase now.
        w.flush()
        assert fo.tell() == file_size_history[-1]

    # Read the records to verify they were written correctly.
    new_reader = fastavro.reader(temp_path.open("rb"))
    new_records = list(new_reader)
    assert new_records == records


def test_writer_class_split_files(tmpdir):
    """
    Create 2 Avro files using the Writer class and the default sync_interval
    setting. We write to one file until the Writer automatically flushes, then
    write more records to the other file. Verify that the two files together
    contain all the records that were written.

    This simulates a real-world use case where a large Avro data set is split
    into files of approximately the same size.
    """
    schema = {
        "type": "record",
        "name": "Test",
        "namespace": "test",
        "fields": [{"name": "field", "type": {"type": "string"}}],
    }
    records = []

    def _append_record(writer_):
        record = {"field": f"test{len(records)}"}
        records.append(record)
        writer_.write(record)

    temp_paths = [
        tmpdir.join("test_writer_class1.avro"),
        tmpdir.join("test_writer_class2.avro"),
    ]
    interim_record_counts = []

    # First file: Write records until block_count goes back to 0 for the second
    # time.
    with temp_paths[0].open("wb") as fo:
        w = Writer(fo, schema, codec="deflate")
        _append_record(w)
        while w.block_count > 0:
            _append_record(w)
        _append_record(w)
        while w.block_count > 0:
            _append_record(w)
        w.flush()
    interim_record_counts.append(len(records))

    # Second file: 100 records
    with temp_paths[1].open("wb") as fo:
        w = Writer(fo, schema, codec="deflate")
        for i in range(100):
            _append_record(w)
        w.flush()
    interim_record_counts.append(len(records))

    assert interim_record_counts[1] == interim_record_counts[0] + 100

    # Read the records to verify they were written correctly.
    new_records = []
    new_interim_record_counts = []
    for temp_path in temp_paths:
        new_reader = fastavro.reader(temp_path.open("rb"))
        new_records += list(new_reader)
        new_interim_record_counts.append(len(new_records))
    assert new_records == records
    assert interim_record_counts == new_interim_record_counts


def test_union_records():
    #
    schema = {
        "name": "test_name",
        "namespace": "test",
        "type": "record",
        "fields": [
            {
                "name": "val",
                "type": [
                    {
                        "name": "a",
                        "namespace": "common",
                        "type": "record",
                        "fields": [
                            {"name": "x", "type": "int"},
                            {"name": "y", "type": "int"},
                        ],
                    },
                    {
                        "name": "b",
                        "namespace": "common",
                        "type": "record",
                        "fields": [
                            {"name": "x", "type": "int"},
                            {"name": "y", "type": "int"},
                            {"name": "z", "type": ["null", "int"]},
                        ],
                    },
                ],
            }
        ],
    }

    data = [
        {
            "val": {
                "x": 3,
                "y": 4,
                "z": 5,
            }
        }
    ]

    assert data == roundtrip(schema, data)


def test_ordered_dict_record():
    """
    Write an Avro record using an OrderedDict and read it back. This tests for
    a bug where dict was supported but not dict-like types.
    """
    schema = {
        "type": "record",
        "name": "Test",
        "namespace": "test",
        "fields": [{"name": "field", "type": {"type": "string"}}],
    }

    record = OrderedDict()
    record["field"] = "foobar"
    records = [record]

    assert records == roundtrip(schema, records)


def test_ordered_dict_map():
    """
    Write an Avro record containing a map field stored in an OrderedDict, then
    read it back. This tests for a bug where dict was supported but not
    dict-like types.
    """
    schema = {
        "type": "record",
        "name": "test_ordered_dict_map",
        "fields": [
            {
                "name": "test",
                "type": {"type": "map", "values": ["string", "int"]},
            }
        ],
    }

    map_ = OrderedDict()
    map_["foo"] = 1
    records = [{"test": map_}]

    assert records == roundtrip(schema, records)


@pytest.mark.skipif(
    not hasattr(_writer, "CYTHON_MODULE"), reason="Cython-specific test"
)
def test_regular_vs_ordered_dict_record_typeerror():
    """
    Tests a corner case where bad data in a dict record causes a TypeError. The
    specific failure lines in the backtrace should be different with dict vs
    OrderedDict, indicating the expected path was taken through the code.
    """
    schema = {
        "type": "record",
        "name": "Test",
        "namespace": "test",
        "fields": [{"name": "field", "type": {"type": "int"}}],
    }

    # Test with two different bad records. One is a regular dict, and the other
    # is an OrderedDict. Both have a bad value (string where the schema
    # declares an int).
    test_records = [{"field": "foobar"}]
    record = OrderedDict()
    record["field"] = "foobar"
    test_records.append(record)

    expected_write_record_stack_traces = [
        # For the regular dict, fails by reraising an error accessing
        # 'd_datum', a variable that only gets a value if the record is an
        # actual dict.
        ["write_data(fo, d_datum_value, field_type, named_schemas, name, options)"],
        # For the OrderedDict, fails directly when accessing 'datum', the
        # variable that is used if the record is *not* an actual dict.
        ["write_data(fo, datum_value, field_type, named_schemas, name, options)"],
    ]

    for test_record, expected_write_record_stack_trace in zip(
        test_records, expected_write_record_stack_traces
    ):
        new_file = BytesIO()
        records = [test_record]
        try:
            fastavro.writer(new_file, schema, records)
            assert False, "Should've raised TypeError"
        except TypeError:
            _, _, tb = sys.exc_info()
            stack = traceback.extract_tb(tb)
            filtered_stack = [frame[3] for frame in stack if "write_record" in frame[2]]
            assert filtered_stack == expected_write_record_stack_trace


@pytest.mark.skipif(
    not hasattr(_writer, "CYTHON_MODULE"), reason="Cython-specific test"
)
def test_regular_vs_ordered_dict_map_typeerror():
    """
    Tests a corner case where bad data in a dict map causes a TypeError. The
    specific failure lines in the backtrace should be different with dict vs
    OrderedDict, indicating the expected path was taken through the code.
    """
    schema = {
        "type": "record",
        "name": "test_regular_vs_ordered_dict_map_typeerror",
        "fields": [
            {
                "name": "test",
                "type": {"type": "map", "values": "int"},
            }
        ],
    }

    # Test with two different bad records. One is a regular dict, and the other
    # is an OrderedDict. Both have a bad value (string where the schema
    # declares an int).
    test_records = [{"test": {"foo": "bar"}}]
    map_ = OrderedDict()
    map_["foo"] = "bar"
    test_records.append({"test": map_})

    filtered_stacks = []
    for test_record in test_records:
        new_file = BytesIO()
        records = [test_record]
        try:
            fastavro.writer(new_file, schema, records)
            assert False, "Should've raised TypeError"
        except TypeError:
            _, _, tb = sys.exc_info()
            stack = traceback.extract_tb(tb)
            filtered_stack = [frame[1] for frame in stack if "write_map" in frame[2]]
            filtered_stacks.append(filtered_stack)

    # Because of the special-case code for dicts, the two stack traces should
    # be different, indicating the exception occurred at a different line
    # number.
    assert filtered_stacks[0] != filtered_stacks[1]


def test_write_union_tuple_primitive():
    """
    Test that when we can use tuple style of writing unions
    (see function `write_union` in `_write`) with primitives
     not only with records.
    """

    schema = {
        "name": "test_name",
        "namespace": "test",
        "type": "record",
        "fields": [{"name": "val", "type": ["string", "int"]}],
    }

    data = [
        {"val": ("int", 1)},
        {"val": ("string", "string")},
    ]

    expected_data = [
        {"val": 1},
        {"val": "string"},
    ]

    new_file = BytesIO()
    fastavro.writer(new_file, schema, data)
    new_file.seek(0)

    new_reader = fastavro.reader(new_file)
    new_records = list(new_reader)

    assert new_records == expected_data


def test_doubles_set_to_zero_on_windows():
    """https://github.com/fastavro/fastavro/issues/154"""

    schema = {
        "doc": "A weather reading.",
        "name": "Weather",
        "namespace": "test",
        "type": "record",
        "fields": [
            {"name": "station", "type": "string"},
            {"name": "time", "type": "long"},
            {"name": "temp", "type": "int"},
            {"name": "test_float", "type": "double"},
        ],
    }

    records = [
        {
            "station": "011990-99999",
            "temp": 0,
            "test_float": 0.21334215134123513,
            "time": -714214260,
        },
        {
            "station": "011990-99999",
            "temp": 22,
            "test_float": 0.21334215134123513,
            "time": -714213259,
        },
        {
            "station": "011990-99999",
            "temp": -11,
            "test_float": 0.21334215134123513,
            "time": -714210269,
        },
        {
            "station": "012650-99999",
            "temp": 111,
            "test_float": 0.21334215134123513,
            "time": -714208170,
        },
    ]

    assert records == roundtrip(schema, records)


def test_string_not_treated_as_array():
    """https://github.com/fastavro/fastavro/issues/166"""

    schema = {
        "type": "record",
        "fields": [
            {
                "name": "description",
                "type": ["null", {"type": "array", "items": "string"}, "string"],
            }
        ],
        "name": "description",
        "doc": "A description of the thing.",
    }

    records = [
        {
            "description": "value",
        },
        {"description": ["an", "array"]},
    ]

    assert records == roundtrip(schema, records)


def test_schema_is_custom_dict_type():
    """https://github.com/fastavro/fastavro/issues/168"""

    class CustomDict(dict):
        pass

    schema = {
        "type": "record",
        "fields": [
            {
                "name": "description",
                "type": ["null", {"type": "array", "items": "string"}, "string"],
            }
        ],
        "name": "description",
        "doc": "A description of the thing.",
    }
    other_type_schema = CustomDict(schema)

    record = {
        "description": "value",
    }

    new_file = BytesIO()
    fastavro.schemaless_writer(new_file, schema, record)
    new_file.seek(0)
    new_record = fastavro.schemaless_reader(new_file, other_type_schema)
    assert record == new_record


def test_long_bounds():
    schema = {
        "name": "test_long_bounds",
        "namespace": "test",
        "type": "record",
        "fields": [
            {"name": "time", "type": "long"},
        ],
    }

    records = [
        {"time": (1 << 63) - 1},
        {"time": -(1 << 63)},
    ]

    assert records == roundtrip(schema, records)


def test_py37_runtime_error():
    """On Python 3.7 this test would cause the StopIteration to get raised as
    a RuntimeError.

    See https://www.python.org/dev/peps/pep-0479/
    """
    weather_file = join(data_dir, "weather.avro")

    zip_io = BytesIO()
    with zipfile.ZipFile(zip_io, mode="w") as zio:
        zio.write(weather_file, arcname="weather")

    with zipfile.ZipFile(zip_io) as zio:
        with zio.open("weather") as fo:
            # Need to read fo into a bytes buffer for python versions less
            # than 3.7
            reader = fastavro.reader(BytesIO(fo.read()))
            list(reader)


def test_eof_error():
    schema = {
        "type": "record",
        "name": "test_eof_error",
        "fields": [
            {
                "name": "test",
                "type": "float",
            }
        ],
    }

    new_file = BytesIO()
    record = {"test": 1.234}
    fastavro.schemaless_writer(new_file, schema, record)

    # Back up one byte and truncate
    new_file.seek(-1, 1)
    new_file.truncate()

    new_file.seek(0)
    with pytest.raises(EOFError):
        fastavro.schemaless_reader(new_file, schema)


def test_eof_error_string():
    schema = "string"
    new_file = BytesIO()
    fastavro.schemaless_writer(new_file, schema, "1234567890")

    # Back up one byte and truncate
    new_file.seek(-1, 1)
    new_file.truncate()

    new_file.seek(0)
    with pytest.raises(EOFError):
        fastavro.schemaless_reader(new_file, schema)


def test_eof_error_fixed():
    schema = {"type": "fixed", "size": 10, "name": "test"}
    new_file = BytesIO()
    fastavro.schemaless_writer(new_file, schema, b"1234567890")

    # Back up one byte and truncate
    new_file.seek(-1, 1)
    new_file.truncate()

    new_file.seek(0)
    with pytest.raises(EOFError):
        fastavro.schemaless_reader(new_file, schema)


def test_eof_error_bytes():
    schema = "bytes"
    new_file = BytesIO()
    fastavro.schemaless_writer(new_file, schema, b"1234567890")

    # Back up one byte and truncate
    new_file.seek(-1, 1)
    new_file.truncate()

    new_file.seek(0)
    with pytest.raises(EOFError):
        fastavro.schemaless_reader(new_file, schema)


def test_write_union_tuple_uses_namespaced_name():
    """
    Test that we must use the fully namespaced name when we are using the tuple
    style of writing unions

    https://github.com/fastavro/fastavro/issues/155
    """

    schema = {
        "name": "test_name",
        "namespace": "test",
        "type": "record",
        "fields": [
            {
                "name": "val",
                "type": [
                    {
                        "name": "A",
                        "namespace": "test",
                        "type": "record",
                        "fields": [
                            {"name": "field", "type": "int"},
                        ],
                    },
                    {
                        "name": "B",
                        "namespace": "test",
                        "type": "record",
                        "fields": [
                            {"name": "field", "type": "string"},
                        ],
                    },
                ],
            }
        ],
    }

    expected_data = [
        {"val": {"field": 1}},
        {"val": {"field": "string"}},
    ]

    data = [
        {"val": ("A", {"field": 1})},
        {"val": ("B", {"field": "string"})},
    ]

    # The given data doesn't use the namespaced name and should fail
    with pytest.raises(ValueError):
        assert expected_data == roundtrip(schema, data)

    data = [
        {"val": ("test.A", {"field": 1})},
        {"val": ("test.B", {"field": "string"})},
    ]

    # This passes because it uses the namespaced name
    assert expected_data == roundtrip(schema, data)


def test_passing_same_schema_to_reader():
    """https://github.com/fastavro/fastavro/issues/244"""
    schema = {
        "namespace": "test.avro.training",
        "name": "SomeMessage",
        "type": "record",
        "fields": [
            {
                "name": "is_error",
                "type": "boolean",
                "default": False,
            },
            {
                "name": "outcome",
                "type": [
                    "SomeMessage",
                    {
                        "type": "record",
                        "name": "ErrorRecord",
                        "fields": [
                            {
                                "name": "errors",
                                "type": {"type": "map", "values": "string"},
                                "doc": "doc",
                            }
                        ],
                    },
                ],
            },
        ],
    }

    records = [
        {
            "is_error": True,
            "outcome": {
                "errors": {"field_1": "some_message", "field_2": "some_other_message"}
            },
        }
    ]

    assert records == roundtrip(schema, records, reader_schema=schema)


def test_helpful_error_when_a_single_record_is_passed_to_writer():
    """https://github.com/fastavro/fastavro/issues/254"""
    schema = {
        "namespace": "namespace",
        "name": "name",
        "type": "record",
        "fields": [
            {
                "name": "is_error",
                "type": "boolean",
            }
        ],
    }

    record = {
        "is_error": True,
    }

    new_file = BytesIO()
    with pytest.raises(ValueError, match="argument should be an iterable, not dict"):
        fastavro.writer(new_file, schema, record)


def test_embedded_records_get_namespaced_correctly():
    schema = {
        "namespace": "test",
        "name": "OuterName",
        "type": "record",
        "fields": [
            {
                "name": "data",
                "type": [
                    {
                        "type": "record",
                        "name": "UUID",
                        "fields": [{"name": "uuid", "type": "string"}],
                    },
                    {
                        "type": "record",
                        "name": "Abstract",
                        "fields": [
                            {
                                "name": "uuid",
                                "type": "UUID",
                            }
                        ],
                    },
                    {
                        "type": "record",
                        "name": "Concrete",
                        "fields": [
                            {"name": "abstract", "type": "Abstract"},
                            {
                                "name": "custom",
                                "type": "string",
                            },
                        ],
                    },
                ],
            }
        ],
    }

    records = [
        {"data": {"abstract": {"uuid": {"uuid": "some_uuid"}}, "custom": "some_string"}}
    ]

    assert records == roundtrip(schema, records)


def test_null_defaults_are_not_used():
    """https://github.com/fastavro/fastavro/issues/272"""
    schema = [
        {
            "type": "record",
            "name": "A",
            "fields": [{"name": "foo", "type": ["string", "null"]}],
        },
        {
            "type": "record",
            "name": "B",
            "fields": [{"name": "bar", "type": ["string", "null"]}],
        },
        {
            "type": "record",
            "name": "AOrB",
            "fields": [{"name": "entity", "type": ["A", "B"]}],
        },
    ]

    datum_to_read = {"entity": {"foo": "this is an instance of schema A"}}

    assert [datum_to_read] == roundtrip(schema, [datum_to_read])


def test_union_schema_ignores_extra_fields():
    """https://github.com/fastavro/fastavro/issues/274"""
    schema = [
        {"type": "record", "name": "A", "fields": [{"name": "name", "type": "string"}]},
        {
            "type": "record",
            "name": "B",
            "fields": [{"name": "other_name", "type": "string"}],
        },
    ]

    records = [{"name": "abc", "other": "asd"}]

    assert [{"name": "abc"}] == roundtrip(schema, records)


def test_appending_records(tmpdir):
    """https://github.com/fastavro/fastavro/issues/276"""
    schema = {
        "type": "record",
        "name": "test_appending_records",
        "fields": [
            {
                "name": "field",
                "type": "string",
            }
        ],
    }

    test_file = str(tmpdir.join("test.avro"))

    with open(test_file, "wb") as new_file:
        fastavro.writer(new_file, schema, [{"field": "foo"}])

    with open(test_file, "a+b") as new_file:
        fastavro.writer(new_file, schema, [{"field": "bar"}])

    with open(test_file, "rb") as new_file:
        reader = fastavro.reader(new_file)
        new_records = list(reader)

    assert new_records == [{"field": "foo"}, {"field": "bar"}]


def test_appending_records_with_io_stream():
    """https://github.com/fastavro/fastavro/issues/276"""
    schema = {
        "type": "record",
        "name": "test_appending_records_with_io_stream",
        "fields": [
            {
                "name": "field",
                "type": "string",
            }
        ],
    }

    stream = BytesIO()

    fastavro.writer(stream, schema, [{"field": "foo"}])

    # Should be able to append to the existing stream
    fastavro.writer(stream, schema, [{"field": "bar"}])

    stream.seek(0)
    reader = fastavro.reader(stream)
    new_records = list(reader)

    assert new_records == [{"field": "foo"}, {"field": "bar"}]

    # If we seek to the beginning and write, it will be treated like a brand
    # new file
    stream.seek(0)
    fastavro.writer(stream, schema, [{"field": "abcdefghijklmnopqrstuvwxyz"}])

    stream.seek(0)
    reader = fastavro.reader(stream)
    new_records = list(reader)

    assert new_records == [{"field": "abcdefghijklmnopqrstuvwxyz"}]


def test_appending_records_wrong_mode_fails(tmpdir):
    """https://github.com/fastavro/fastavro/issues/276"""
    schema = {
        "type": "record",
        "name": "test_appending_records_wrong_mode_fails",
        "fields": [
            {
                "name": "field",
                "type": "string",
            }
        ],
    }

    test_file = str(tmpdir.join("test.avro"))

    with open(test_file, "wb") as new_file:
        fastavro.writer(new_file, schema, [{"field": "foo"}])

    with open(test_file, "ab") as new_file:
        with pytest.raises(
            ValueError, match=r"you must use the 'a\+' mode, not just 'a'"
        ):
            fastavro.writer(new_file, schema, [{"field": "bar"}])


def test_appending_records_different_schema_works(tmpdir):
    """https://github.com/fastavro/fastavro/issues/276"""
    schema = {
        "type": "record",
        "name": "test_appending_records_different_schema_fails",
        "fields": [{"name": "field", "type": "string"}],
    }

    test_file = str(tmpdir.join("test.avro"))

    with open(test_file, "wb") as new_file:
        fastavro.writer(new_file, schema, [{"field": "foo"}])

    different_schema = {
        "type": "record",
        "name": "test_appending_records",
        "fields": [{"name": "field", "type": "int"}],
    }

    with open(test_file, "a+b") as new_file:
        fastavro.writer(new_file, different_schema, [{"field": "bar"}])


def test_appending_records_different_schema_works_2(tmpdir):
    """https://github.com/fastavro/fastavro/issues/276"""
    schema = {
        "type": "record",
        "name": "test_appending_records_different_schema_fails",
        "fields": [
            {"name": "field", "type": "string"},
            {
                "name": "field2",
                "type": {
                    "type": "record",
                    "name": "subrecord",
                    "fields": [{"name": "subfield", "type": "string"}],
                },
            },
            {"name": "field3", "type": "subrecord"},
        ],
    }

    test_file = str(tmpdir.join("test.avro"))

    with open(test_file, "wb") as new_file:
        fastavro.writer(
            new_file,
            schema,
            [
                {
                    "field": "foo",
                    "field2": {"subfield": "foo2"},
                    "field3": {"subfield": "foo3"},
                }
            ],
        )

    different_schema = {
        "type": "record",
        "name": "test_appending_records",
        "fields": [{"name": "field", "type": "int"}],
    }

    with open(test_file, "a+b") as new_file:
        fastavro.writer(
            new_file,
            different_schema,
            [
                {
                    "field": "bar",
                    "field2": {"subfield": "bar2"},
                    "field3": {"subfield": "bar3"},
                }
            ],
        )


def test_appending_records_null_schema_works(tmpdir):
    """https://github.com/fastavro/fastavro/issues/422"""
    schema = {
        "type": "record",
        "name": "test_appending_records_different_schema_fails",
        "fields": [{"name": "field", "type": "string"}],
    }

    test_file = str(tmpdir.join("test.avro"))

    with open(test_file, "wb") as new_file:
        fastavro.writer(new_file, schema, [{"field": "foo"}])

    with open(test_file, "a+b") as new_file:
        fastavro.writer(new_file, None, [{"field": "bar"}])


def test_user_specified_sync():
    """https://github.com/fastavro/fastavro/issues/300"""
    schema = {"type": "record", "name": "test_user_specified_sync", "fields": []}

    file1 = BytesIO()
    file2 = BytesIO()

    records = [{}]

    fastavro.writer(file1, schema, records, sync_marker=b"16bytesyncmarker")
    fastavro.writer(file2, schema, records, sync_marker=b"16bytesyncmarker")

    assert file1.getvalue() == file2.getvalue()


def test_order_of_values_in_map():
    """https://github.com/fastavro/fastavro/issues/303"""
    schema = {
        "doc": "A weather reading.",
        "name": "Weather",
        "namespace": "test",
        "type": "record",
        "fields": [
            {
                "name": "metadata",
                "type": {
                    "type": "map",
                    "values": [
                        {"type": "array", "items": "string"},
                        {"type": "map", "values": ["string"]},
                    ],
                },
            }
        ],
    }
    parsed_schema = fastavro.parse_schema(schema)

    records = [{"metadata": {"map1": {"map2": "str"}}}]

    assert records == roundtrip(parsed_schema, records)


def test_reader_schema_attributes_throws_deprecation():
    """https://github.com/fastavro/fastavro/issues/246"""
    schema = {
        "type": "record",
        "name": "test_reader_schema_attributes_throws_deprecation",
        "fields": [],
    }

    stream = BytesIO()

    fastavro.writer(stream, schema, [{}])
    stream.seek(0)

    reader = fastavro.reader(stream)
    with pytest.warns(DeprecationWarning):
        reader.schema


def test_writer_schema_always_read():
    """https://github.com/fastavro/fastavro/issues/312"""
    schema = {
        "type": "record",
        "name": "Outer",
        "fields": [
            {
                "name": "item",
                "type": [
                    {
                        "type": "record",
                        "name": "Inner1",
                        "fields": [
                            {
                                "name": "id",
                                "type": {
                                    "type": "record",
                                    "name": "UUID",
                                    "fields": [{"name": "id", "type": "string"}],
                                },
                                "default": {"id": ""},
                            },
                            {"name": "description", "type": "string"},
                            {"name": "size", "type": "int"},
                        ],
                    },
                    {
                        "type": "record",
                        "name": "Inner2",
                        "fields": [
                            {"name": "id", "type": "UUID", "default": {"id": ""}},
                            {"name": "name", "type": "string"},
                            {"name": "age", "type": "long"},
                        ],
                    },
                ],
            }
        ],
    }

    records = [
        {"item": {"description": "test", "size": 1}},
        {"item": {"id": {"id": "#1"}, "name": "foobar", "age": 12}},
    ]

    file = BytesIO()

    fastavro.writer(file, fastavro.parse_schema(schema), records)
    file.seek(0)

    # This should not raise a KeyError
    fastavro.reader(file)


def test_hint_is_not_written_to_the_file():
    """The __fastavro_parsed hint should not be written to the avrofile"""
    schema = {
        "type": "record",
        "name": "test_hint_is_not_written_to_the_file",
        "fields": [],
    }

    parsed_schema = fastavro.parse_schema(schema)

    # It should get added when parsing
    assert "__fastavro_parsed" in parsed_schema

    stream = BytesIO()
    fastavro.writer(stream, parsed_schema, [{}])
    stream.seek(0)

    reader = fastavro.reader(stream)
    # By the time it is in the file, it should not
    assert "__fastavro_parsed" not in reader._schema


def test_hint_is_not_written_to_the_file_list_schema():
    """The __fastavro_parsed hint should not be written to the avrofile"""
    schema = [
        {
            "type": "record",
            "name": "test_hint_is_not_written_to_the_file_list_schema_1",
            "fields": [],
        },
        {
            "type": "record",
            "name": "test_hint_is_not_written_to_the_file_list_schema_2",
            "fields": [],
        },
    ]

    parsed_schema = fastavro.parse_schema(schema)

    # It should get added when parsing
    assert all(("__fastavro_parsed" in s for s in parsed_schema))

    stream = BytesIO()
    fastavro.writer(stream, parsed_schema, [{}])
    stream.seek(0)

    reader = fastavro.reader(stream)
    # By the time it is in the file, it should not
    assert all(("__fastavro_parsed" not in s for s in reader._schema))


def test_more_null_union_issues():
    """https://github.com/fastavro/fastavro/issues/336"""
    schema = {
        "type": "record",
        "name": "test",
        "fields": [
            {"name": "name", "type": "string"},
            {
                "name": "address",
                "type": [
                    "null",
                    {
                        "type": "array",
                        "items": {
                            "type": "record",
                            "name": "whatever",
                            "fields": [
                                {"name": "street", "type": ["null", "string"]},
                                {"name": "zip", "type": ["null", "string"]},
                            ],
                        },
                    },
                ],
            },
        ],
    }

    records = [
        {"name": "name1", "address": [{"street": "22st"}]},
        {"name": "name2"},
    ]

    expected = [
        {"name": "name1", "address": [{"street": "22st", "zip": None}]},
        {"name": "name2", "address": None},
    ]

    assert expected == roundtrip(schema, records)


def test_logical_type_in_union():
    schema = {
        "type": "record",
        "name": "test_logical_type_in_union",
        "fields": [
            {"name": "item", "type": ["null", {"type": "int", "logicalType": "date"}]}
        ],
    }

    records = [{"item": None}, {"item": "2019-05-06"}]

    expected = [{"item": None}, {"item": datetime.date(2019, 5, 6)}]

    assert expected == roundtrip(schema, records)


def test_named_schema_with_logical_type_in_union():
    schema = [
        {
            "name": "named_schema_with_logical_type",
            "namespace": "com.example",
            "type": "record",
            "fields": [
                {"name": "item", "type": {"type": "int", "logicalType": "date"}}
            ],
        },
        {
            "type": "record",
            "name": "test_named_schema_with_logical_type",
            "fields": [
                {
                    "name": "item",
                    "type": ["null", "com.example.named_schema_with_logical_type"],
                }
            ],
        },
    ]

    records = [{"item": None}, {"item": {"item": "2019-05-06"}}]

    expected = [{"item": None}, {"item": {"item": datetime.date(2019, 5, 6)}}]

    assert expected == roundtrip(schema, records)


def test_return_record_name_with_named_type_in_union():
    schema = {
        "type": "record",
        "name": "my_record",
        "fields": [
            {
                "name": "my_1st_union",
                "type": [
                    {
                        "name": "foo",
                        "type": "record",
                        "fields": [{"name": "some_field", "type": "int"}],
                    },
                    {
                        "name": "bar",
                        "type": "record",
                        "fields": [{"name": "some_field", "type": "int"}],
                    },
                ],
            },
            {"name": "my_2nd_union", "type": ["foo", "bar"]},
        ],
    }

    records = [
        {
            "my_1st_union": ("foo", {"some_field": 1}),
            "my_2nd_union": ("bar", {"some_field": 2}),
        }
    ]

    rt_records = roundtrip(
        fastavro.parse_schema(schema), records, return_record_name=True
    )
    assert records == rt_records


def test_return_record_with_named_type_in_union():
    """https://github.com/fastavro/fastavro/issues/625"""
    schema = {
        "type": "record",
        "name": "my_record",
        "fields": [
            {
                "name": "my_union",
                "type": [
                    "null",
                    {
                        "name": "bar",
                        "type": "record",
                        "fields": [{"name": "some_field", "type": "int"}],
                    },
                ],
            }
        ],
    }

    records = [{"my_union": None}, {"my_union": {"some_field": 2}}]

    rt_records = roundtrip(
        fastavro.parse_schema(schema),
        records,
        return_record_name=True,
        return_record_name_override=True,
    )
    assert records == rt_records


def test_return_record_name_with_named_type_and_null_in_union():
    """https://github.com/fastavro/fastavro/issues/625"""
    schema = {
        "type": "record",
        "name": "my_record",
        "fields": [
            {
                "name": "my_union",
                "type": [
                    "null",
                    {
                        "name": "foo",
                        "type": "record",
                        "fields": [{"name": "some_field", "type": "int"}],
                    },
                    {
                        "name": "bar",
                        "type": "record",
                        "fields": [{"name": "some_field", "type": "int"}],
                    },
                ],
            }
        ],
    }

    records = [
        {"my_union": None},
        {"my_union": ("foo", {"some_field": 1})},
        {"my_union": ("bar", {"some_field": 2})},
    ]

    rt_records = roundtrip(
        fastavro.parse_schema(schema),
        records,
        return_record_name=True,
        return_record_name_override=True,
    )
    assert records == rt_records


def test_return_record_name_with_named_type_and_null_in_union2():
    """https://github.com/fastavro/fastavro/issues/625"""
    schema = {
        "type": "record",
        "name": "my_record",
        "fields": [
            {
                "name": "foo",
                "type": {
                    "type": "record",
                    "name": "Foo",
                    "fields": [{"name": "subfoo", "type": "string"}],
                },
            },
            {
                "name": "my_union",
                "type": [
                    "null",
                    "Foo",
                    {
                        "name": "bar",
                        "type": "record",
                        "fields": [{"name": "some_field", "type": "int"}],
                    },
                ],
            },
        ],
    }

    records = [
        {"foo": {"subfoo": "subfoo"}, "my_union": None},
        {"foo": {"subfoo": "subfoo"}, "my_union": ("Foo", {"subfoo": "1"})},
        {"foo": {"subfoo": "subfoo"}, "my_union": ("bar", {"some_field": 2})},
    ]

    rt_records = roundtrip(
        fastavro.parse_schema(schema),
        records,
        return_record_name=True,
        return_record_name_override=True,
    )
    assert records == rt_records


def test_return_record_with_multiple_simple_types_and_null_in_union():
    """https://github.com/fastavro/fastavro/issues/625"""
    schema = {
        "type": "record",
        "name": "my_record",
        "fields": [
            {
                "name": "my_union",
                "type": [
                    "null",
                    "string",
                    "int",
                    {
                        "name": "foo",
                        "type": "record",
                        "fields": [{"name": "some_field", "type": "int"}],
                    },
                ],
            }
        ],
    }

    records = [
        {"my_union": None},
        {"my_union": "3"},
        {"my_union": 3},
        {"my_union": {"some_field": 2}},
    ]

    rt_records = roundtrip(
        fastavro.parse_schema(schema),
        records,
        return_record_name=True,
        return_record_name_override=True,
    )
    assert records == rt_records


def test_return_record_with_multiple_dict_types_and_null_in_union():
    schema = {
        "type": "record",
        "name": "my_record",
        "fields": [
            {
                "name": "my_union",
                "type": [
                    "null",
                    "int",
                    {
                        "type": "enum",
                        "name": "my_enum",
                        "symbols": ["FOO", "BAR"],
                    },
                    {
                        "name": "foo",
                        "type": "record",
                        "fields": [{"name": "some_field", "type": "int"}],
                    },
                ],
            }
        ],
    }

    records = [
        {"my_union": None},
        {"my_union": 3},
        {"my_union": "FOO"},
        {"my_union": {"some_field": 2}},
    ]

    rt_records = roundtrip(
        fastavro.parse_schema(schema),
        records,
        return_record_name=True,
        return_record_name_override=True,
    )
    assert records == rt_records


def test_enum_named_type():
    """https://github.com/fastavro/fastavro/issues/450"""
    schema = {
        "type": "record",
        "name": "test_enum_named_type",
        "fields": [
            {
                "name": "test1",
                "type": {
                    "type": "enum",
                    "name": "my_enum",
                    "symbols": ["FOO", "BAR"],
                },
            },
            {
                "name": "test2",
                "type": "my_enum",
            },
        ],
    }

    records = [{"test1": "FOO", "test2": "BAR"}]
    parsed_schema = fastavro.parse_schema(schema)
    assert records == roundtrip(parsed_schema, records)


def test_fixed_named_type():
    """https://github.com/fastavro/fastavro/issues/450"""
    schema = {
        "type": "record",
        "name": "test_fixed_named_type",
        "fields": [
            {
                "name": "test1",
                "type": {
                    "type": "fixed",
                    "name": "my_fixed",
                    "size": 4,
                },
            },
            {
                "name": "test2",
                "type": "my_fixed",
            },
        ],
    }

    records = [{"test1": b"1234", "test2": b"4321"}]
    parsed_schema = fastavro.parse_schema(schema)
    assert records == roundtrip(parsed_schema, records)


def test_record_named_type():
    """https://github.com/fastavro/fastavro/issues/450"""
    schema = {
        "type": "record",
        "name": "test_record_named_type",
        "fields": [
            {
                "name": "test1",
                "type": {
                    "type": "record",
                    "name": "my_record",
                    "fields": [
                        {
                            "name": "field1",
                            "type": "string",
                        }
                    ],
                },
            },
            {
                "name": "test2",
                "type": "my_record",
            },
        ],
    }

    records = [{"test1": {"field1": "foo"}, "test2": {"field1": "bar"}}]
    parsed_schema = fastavro.parse_schema(schema)
    assert records == roundtrip(parsed_schema, records)


def test_write_required_field_name():
    """https://github.com/fastavro/fastavro/issues/439

    Test that when a TypeError is raised the fieldname is
    included in exception message allowing to figure out
    quickly what column value mismatches the schema field type.
    Useful for Null/None values in required fields also.
    """

    schema = {
        "name": "test_required",
        "namespace": "test",
        "type": "record",
        "fields": [
            {"name": "person", "type": "string"},
            {"name": "age", "type": "int"},
        ],
    }

    data = [
        {"person": "James", "age": 34},
        {"person": "Anthony", "age": None},
    ]

    new_file = BytesIO()
    with pytest.raises(TypeError, match="on field age"):
        fastavro.writer(new_file, schema, data)


def test_write_mismatched_field_type():
    """https://github.com/fastavro/fastavro/issues/439

    Test that when a ValueError is raised the fieldname is
    included in exception message allowing to figure out
    quickly what column mismatches the schema field type.
    """

    schema = {
        "name": "test_required",
        "namespace": "test",
        "type": "record",
        "fields": [
            {"name": "person", "type": "string"},
            {"name": "age", "type": ["null", "int"]},
        ],
    }

    data = [
        {"person": "James", "age": 34},
        {"person": "Anthony", "age": "26"},
    ]

    new_file = BytesIO()
    with pytest.raises(ValueError, match="on field age"):
        fastavro.writer(new_file, schema, data)


def test_reading_with_subschema():
    """https://github.com/fastavro/fastavro/issues/503"""
    writer_schema = {
        "type": "record",
        "name": "Test",
        "namespace": "test",
        "fields": [
            {"name": "null", "type": "null"},
            {"name": "boolean", "type": "boolean"},
            {"name": "string", "type": "string"},
            {"name": "bytes", "type": "bytes"},
            {"name": "int", "type": "int"},
            {"name": "long", "type": "long"},
            {"name": "float", "type": "float"},
            {"name": "double", "type": "double"},
            {
                "name": "fixed",
                "type": {"type": "fixed", "name": "fixed_field", "size": 5},
            },
            {
                "name": "union",
                "type": [
                    "null",
                    "int",
                    {
                        "type": "record",
                        "name": "union_record",
                        "fields": [{"name": "union_record_field", "type": "string"}],
                    },
                ],
            },
            {
                "name": "enum",
                "type": {
                    "type": "enum",
                    "name": "enum_field",
                    "symbols": ["FOO", "BAR"],
                },
            },
            {"name": "array", "type": {"type": "array", "items": "string"}},
            {"name": "map", "type": {"type": "map", "values": "int"}},
            {
                "name": "record",
                "type": {
                    "type": "record",
                    "name": "subrecord",
                    "fields": [{"name": "sub_int", "type": "int"}],
                },
            },
        ],
    }

    reader_schema = {
        "type": "record",
        "name": "Test",
        "namespace": "test",
        "fields": [
            {"name": "string", "type": "string"},
            {"name": "double", "type": "double"},
            {
                "name": "enum",
                "type": {
                    "type": "enum",
                    "name": "enum_field",
                    "symbols": ["FOO", "BAR"],
                },
            },
        ],
    }

    records = [
        {
            "null": None,
            "boolean": True,
            "string": "foo",
            "bytes": b"\xe2\x99\xa5",
            "int": 1,
            "long": 1 << 33,
            "float": 2.2,
            "double": 3.3,
            "fixed": b"\x61\x61\x61\x61\x61",
            "union": None,
            "enum": "BAR",
            "array": ["a", "b"],
            "map": {"c": 1, "d": 2},
            "record": {
                "sub_int": 123,
            },
        },
    ]

    roundtrip_records = roundtrip(writer_schema, records, reader_schema=reader_schema)
    assert roundtrip_records == [{"string": "foo", "double": 3.3, "enum": "BAR"}]


@pytest.mark.skipif(
    not hasattr(_reader, "CYTHON_MODULE"), reason="Only works using cython module"
)
def test_reading_with_skip_using_cython():
    """https://github.com/fastavro/fastavro/issues/503"""
    writer_schema = {
        "type": "record",
        "name": "Test",
        "namespace": "test",
        "fields": [
            {"name": "null", "type": "null"},
            {"name": "boolean", "type": "boolean"},
            {"name": "string", "type": "string"},
            {"name": "bytes", "type": "bytes"},
            {"name": "int", "type": "int"},
            {"name": "long", "type": "long"},
            {"name": "float", "type": "float"},
            {"name": "double", "type": "double"},
            {
                "name": "fixed",
                "type": {"type": "fixed", "name": "fixed_field", "size": 5},
            },
            {
                "name": "union",
                "type": [
                    "null",
                    "int",
                    {
                        "type": "record",
                        "name": "union_record",
                        "fields": [{"name": "union_record_field", "type": "string"}],
                    },
                ],
            },
            {
                "name": "enum",
                "type": {
                    "type": "enum",
                    "name": "enum_field",
                    "symbols": ["FOO", "BAR"],
                },
            },
            {
                "name": "enum2",
                "type": {
                    "type": "enum",
                    "name": "enum_field2",
                    "symbols": ["BAZ", "BAZBAZ"],
                },
            },
            {"name": "array", "type": {"type": "array", "items": "string"}},
            {"name": "map", "type": {"type": "map", "values": "int"}},
            {
                "name": "record",
                "type": {
                    "type": "record",
                    "name": "subrecord",
                    "fields": [{"name": "sub_int", "type": "int"}],
                },
            },
            {"name": "record2", "type": "test.subrecord"},
        ],
    }

    reader_schema = {
        "type": "record",
        "name": "Test",
        "namespace": "test",
        "fields": [
            {"name": "string", "type": "string"},
            {"name": "double", "type": "double"},
            {
                "name": "enum",
                "type": {
                    "type": "enum",
                    "name": "enum_field",
                    "symbols": ["FOO", "BAR"],
                },
            },
        ],
    }

    records = [
        {
            "null": None,
            "boolean": True,
            "string": "foo",
            "bytes": b"\xe2\x99\xa5",
            "int": 1,
            "long": 1 << 33,
            "float": 2.2,
            "double": 3.3,
            "fixed": b"\x61\x61\x61\x61\x61",
            "union": None,
            "enum": "BAR",
            "enum2": "BAZBAZ",
            "array": ["a", "b"],
            "map": {"c": 1, "d": 2},
            "record": {"sub_int": 123},
            "record2": {"sub_int": 321},
        },
    ]

    named_schemas = {}
    parsed_writer_schema = fastavro.parse_schema(writer_schema, named_schemas)
    roundtrip_records = roundtrip(
        parsed_writer_schema, records, reader_schema=reader_schema
    )

    new_file = BytesIO()
    fastavro.writer(new_file, parsed_writer_schema, records)
    new_file.seek(0)

    skip_record = {}
    _reader.skip_record(new_file, HEADER_SCHEMA, {})

    block_count = _reader.read_long(new_file)
    assert block_count == 1

    # Skip size in bytes of the serialized objects in the block
    _reader.skip_long(new_file)

    _reader.skip_null(new_file)
    _reader.skip_boolean(new_file)
    skip_record["string"] = _reader.read_utf8(new_file)
    _reader.skip_bytes(new_file)
    _reader.skip_int(new_file)
    _reader.skip_long(new_file)
    _reader.skip_float(new_file)
    skip_record["double"] = _reader.read_double(new_file)
    _reader.skip_fixed(new_file, {"type": "fixed", "name": "fixed_field", "size": 5})
    _reader.skip_union(
        new_file,
        [
            "null",
            "int",
            {
                "type": "record",
                "name": "union_record",
                "fields": [{"name": "union_record_field", "type": "string"}],
            },
        ],
        named_schemas,
    )
    skip_record["enum"] = _reader.read_enum(
        new_file,
        {"type": "enum", "name": "enum_field", "symbols": ["FOO", "BAR"]},
        {"type": "enum", "name": "enum_field", "symbols": ["FOO", "BAR"]},
    )
    _reader.skip_enum(new_file)
    _reader.skip_array(new_file, {"type": "array", "items": "string"}, named_schemas)
    _reader.skip_map(new_file, {"type": "map", "values": "int"}, named_schemas)
    _reader.skip_record(
        new_file,
        {
            "type": "record",
            "name": "subrecord",
            "fields": [{"name": "sub_int", "type": "int"}],
        },
        named_schemas,
    )
    _reader.skip_record(
        new_file,
        {
            "type": "record",
            "name": "subrecord",
            "fields": [{"name": "sub_int", "type": "int"}],
        },
        named_schemas,
    )

    assert roundtrip_records == [skip_record]


@pytest.mark.skipif(
    hasattr(_reader, "CYTHON_MODULE"), reason="Only works when not using cython module"
)
def test_reading_with_skip_using_pure_python():
    """https://github.com/fastavro/fastavro/issues/503"""
    writer_schema = {
        "type": "record",
        "name": "Test",
        "namespace": "test",
        "fields": [
            {"name": "null", "type": "null"},
            {"name": "boolean", "type": "boolean"},
            {"name": "string", "type": "string"},
            {"name": "bytes", "type": "bytes"},
            {"name": "int", "type": "int"},
            {"name": "long", "type": "long"},
            {"name": "float", "type": "float"},
            {"name": "double", "type": "double"},
            {
                "name": "fixed",
                "type": {"type": "fixed", "name": "fixed_field", "size": 5},
            },
            {
                "name": "union",
                "type": [
                    "null",
                    "int",
                    {
                        "type": "record",
                        "name": "union_record",
                        "fields": [{"name": "union_record_field", "type": "string"}],
                    },
                ],
            },
            {
                "name": "enum",
                "type": {
                    "type": "enum",
                    "name": "enum_field",
                    "symbols": ["FOO", "BAR"],
                },
            },
            {
                "name": "enum2",
                "type": {
                    "type": "enum",
                    "name": "enum_field2",
                    "symbols": ["BAZ", "BAZBAZ"],
                },
            },
            {"name": "array", "type": {"type": "array", "items": "string"}},
            {"name": "map", "type": {"type": "map", "values": "int"}},
            {
                "name": "record",
                "type": {
                    "type": "record",
                    "name": "subrecord",
                    "fields": [{"name": "sub_int", "type": "int"}],
                },
            },
            {"name": "record2", "type": "test.subrecord"},
        ],
    }

    reader_schema = {
        "type": "record",
        "name": "Test",
        "namespace": "test",
        "fields": [
            {"name": "string", "type": "string"},
            {"name": "double", "type": "double"},
            {
                "name": "enum",
                "type": {
                    "type": "enum",
                    "name": "enum_field",
                    "symbols": ["FOO", "BAR"],
                },
            },
        ],
    }

    records = [
        {
            "null": None,
            "boolean": True,
            "string": "foo",
            "bytes": b"\xe2\x99\xa5",
            "int": 1,
            "long": 1 << 33,
            "float": 2.2,
            "double": 3.3,
            "fixed": b"\x61\x61\x61\x61\x61",
            "union": None,
            "enum": "BAR",
            "enum2": "BAZBAZ",
            "array": ["a", "b"],
            "map": {"c": 1, "d": 2},
            "record": {"sub_int": 123},
            "record2": {"sub_int": 321},
        },
    ]

    named_schemas = {}
    parsed_writer_schema = fastavro.parse_schema(writer_schema, named_schemas)
    roundtrip_records = roundtrip(
        parsed_writer_schema, records, reader_schema=reader_schema
    )

    new_file = BytesIO()
    fastavro.writer(new_file, parsed_writer_schema, records)
    new_file.seek(0)

    skip_record = {}
    decoder = BinaryDecoder(new_file)
    _reader.skip_record(decoder, HEADER_SCHEMA, {})

    block_count = _reader.read_long(decoder)
    assert block_count == 1

    # Skip size in bytes of the serialized objects in the block
    _reader.skip_long(decoder)

    _reader.skip_null(decoder)
    _reader.skip_boolean(decoder)
    skip_record["string"] = _reader.read_utf8(decoder)
    _reader.skip_bytes(decoder)
    _reader.skip_int(decoder)
    _reader.skip_long(decoder)
    _reader.skip_float(decoder)
    skip_record["double"] = _reader.read_double(decoder)
    _reader.skip_fixed(decoder, {"type": "fixed", "name": "fixed_field", "size": 5})
    _reader.skip_union(
        decoder,
        [
            "null",
            "int",
            {
                "type": "record",
                "name": "union_record",
                "fields": [{"name": "union_record_field", "type": "string"}],
            },
        ],
        named_schemas,
    )
    skip_record["enum"] = _reader.read_enum(
        decoder,
        {"type": "enum", "name": "enum_field", "symbols": ["FOO", "BAR"]},
        named_schemas,
        {"type": "enum", "name": "enum_field", "symbols": ["FOO", "BAR"]},
    )
    _reader.skip_enum(
        decoder,
        {"type": "enum", "name": "enum_field2", "symbols": ["BAZ", "BAZBAZ"]},
        {"type": "enum", "name": "enum_field2", "symbols": ["BAZ", "BAZBAZ"]},
    )
    _reader.skip_array(decoder, {"type": "array", "items": "string"}, named_schemas)
    _reader.skip_map(decoder, {"type": "map", "values": "int"}, named_schemas)
    _reader.skip_record(
        decoder,
        {
            "type": "record",
            "name": "subrecord",
            "fields": [{"name": "sub_int", "type": "int"}],
        },
        named_schemas,
    )
    _reader.skip_record(
        decoder,
        {
            "type": "record",
            "name": "subrecord",
            "fields": [{"name": "sub_int", "type": "int"}],
        },
        named_schemas,
    )

    assert roundtrip_records == [skip_record]


def test_tuple_writer_picks_correct_union_path():
    """https://github.com/fastavro/fastavro/issues/509"""
    schema = {
        "type": "record",
        "name": "test_tuple_writer_picks_correct_union_path",
        "fields": [
            {
                "name": "Field",
                "type": [
                    {"type": "map", "values": "string"},
                    {
                        "type": "record",
                        "name": "Record2",
                        "fields": [{"name": "Field", "type": "string"}],
                    },
                ],
            },
        ],
    }

    records = [{"Field": ("Record2", {"Field": "value"})}]
    parsed_schema = fastavro.parse_schema(schema)
    assert records == roundtrip(parsed_schema, records, return_record_name=True)

    records = [{"Field": ("map", {"Field": "value"})}]
    expected_roundtrip_value = [{"Field": {"Field": "value"}}]
    parsed_schema = fastavro.parse_schema(schema)
    assert expected_roundtrip_value == roundtrip(
        parsed_schema, records, return_record_name=True
    )


def test_schema_migration_should_match_name():
    """https://github.com/fastavro/fastavro/issues/512"""
    writer_schema = {
        "type": "record",
        "name": "root",
        "namespace": "example.space",
        "fields": [
            {"name": "id", "type": "string"},
            {
                "name": "data",
                "type": [
                    "null",
                    {
                        "type": "record",
                        "name": "SomeData",
                        "fields": [
                            {"name": "somedata_id", "type": "int"},
                            {"name": "somedata_field", "type": "int"},
                        ],
                    },
                    {
                        "type": "record",
                        "name": "OtherData",
                        "fields": [
                            {"name": "otherdata_id", "type": "int"},
                            {"name": "otherdata_field", "type": "string"},
                        ],
                    },
                ],
            },
        ],
    }

    reader_schema = {
        "type": "record",
        "name": "root",
        "namespace": "example.space",
        "fields": [
            {"name": "id", "type": "string"},
            {
                "name": "data",
                "type": [
                    "null",
                    {
                        "type": "record",
                        "name": "SomeData",
                        "fields": [
                            {"name": "somedata_id", "type": "int"},
                            {"name": "somedata_field", "type": "int"},
                        ],
                    },
                    {
                        "type": "record",
                        "name": "OtherData",
                        "fields": [
                            {"name": "otherdata_id", "type": "int"},
                            {"name": "otherdata_field", "type": "string"},
                            # Fully compatible change
                            {
                                "name": "newdata_field",
                                "type": "string",
                                "default": "new stuff",
                            },
                        ],
                    },
                ],
            },
        ],
    }

    records = [
        {
            "id": "example_id_1234",
            "data": {"otherdata_id": 1234, "otherdata_field": "some example data"},
        }
    ]

    roundtrip_records = roundtrip(writer_schema, records, reader_schema=reader_schema)

    expected_roundtrip = [
        {
            "id": "example_id_1234",
            "data": {
                "otherdata_id": 1234,
                "otherdata_field": "some example data",
                "newdata_field": "new stuff",
            },
        }
    ]
    assert roundtrip_records == expected_roundtrip


def test_schema_migration_should_fail_with_different_names():
    """https://github.com/fastavro/fastavro/issues/515"""
    writer_schema = {
        "type": "record",
        "name": "A",
        "fields": [{"type": "int", "name": "F"}],
    }

    reader_schema = {
        "type": "record",
        "name": "B",
        "fields": [{"type": "int", "name": "F"}],
    }

    records = [{"F": 1}]

    with pytest.raises(SchemaResolutionError):
        roundtrip(writer_schema, records, reader_schema=reader_schema)


def test_union_of_float_and_double_keeps_precision():
    """https://github.com/fastavro/fastavro/issues/437"""
    schema = ["float", "string", "double"]
    records = [
        1.0,
        1e200,  # Turns into float("+inf") if parsed as 32 bit float
    ]
    parsed_schema = fastavro.parse_schema(schema)
    assert records == roundtrip(parsed_schema, records)


def test_union_of_float_and_no_double():
    """https://github.com/fastavro/fastavro/issues/437"""
    schema = ["float", "string"]
    records = [1.0]
    parsed_schema = fastavro.parse_schema(schema)
    assert records == roundtrip(parsed_schema, records)


def test_error_if_trying_to_write_the_wrong_number_of_bytes():
    """https://github.com/fastavro/fastavro/issues/522"""
    schema = {"type": "fixed", "size": 2, "name": "fixed"}
    parsed_schema = fastavro.parse_schema(schema)

    records = [b"22", b"1", b"22"]
    with pytest.raises(ValueError):
        roundtrip(parsed_schema, records)

    records = [b"22", b"333", b"22"]
    with pytest.raises(ValueError):
        roundtrip(parsed_schema, records)


def test_tuple_writer_picks_correct_union_path_enum():
    """https://github.com/fastavro/fastavro/issues/536"""
    schema = {
        "type": "record",
        "name": "test_tuple_writer_picks_correct_union_path_enum",
        "fields": [
            {
                "name": "Field",
                "type": [
                    {"type": "string"},
                    {"type": "enum", "name": "Enum", "symbols": ["FOO", "BAR"]},
                ],
            },
        ],
    }

    expected_roundtrip_value = [{"Field": "FOO"}]

    records = [{"Field": ("Enum", "FOO")}]
    parsed_schema = fastavro.parse_schema(schema)
    assert expected_roundtrip_value == roundtrip(
        parsed_schema, records, return_record_name=True
    )

    records = [{"Field": ("string", "FOO")}]
    expected_roundtrip_value = [{"Field": "FOO"}]
    parsed_schema = fastavro.parse_schema(schema)
    assert expected_roundtrip_value == roundtrip(
        parsed_schema, records, return_record_name=True
    )


def test_tuple_writer_picks_correct_union_path_fixed():
    """https://github.com/fastavro/fastavro/issues/536"""
    schema = {
        "type": "record",
        "name": "test_tuple_writer_picks_correct_union_path_fixed",
        "fields": [
            {
                "name": "Field",
                "type": [
                    {"type": "bytes"},
                    {"type": "fixed", "name": "Fixed", "size": 4},
                ],
            },
        ],
    }

    expected_roundtrip_value = [{"Field": b"1234"}]

    records = [{"Field": ("Fixed", b"1234")}]
    parsed_schema = fastavro.parse_schema(schema)
    assert expected_roundtrip_value == roundtrip(
        parsed_schema, records, return_record_name=True
    )

    records = [{"Field": ("bytes", b"1234")}]
    expected_roundtrip_value = [{"Field": b"1234"}]
    parsed_schema = fastavro.parse_schema(schema)
    assert expected_roundtrip_value == roundtrip(
        parsed_schema, records, return_record_name=True
    )


@pytest.mark.parametrize(
    "input_records,expected_roundtrip",
    [
        (
            [
                {
                    "my_1st_union": {"some_field": 1, "-type": "foo"},
                    "my_2nd_union": {"some_field": 2, "-type": "foo"},
                }
            ],
            [
                {
                    "my_1st_union": ("foo", {"some_field": 1}),
                    "my_2nd_union": ("foo", {"some_field": 2}),
                }
            ],
        ),
        (
            [
                {
                    "my_1st_union": {"some_field": 3, "-type": "foo"},
                    "my_2nd_union": {"some_field": 4, "-type": "bar"},
                }
            ],
            [
                {
                    "my_1st_union": ("foo", {"some_field": 3}),
                    "my_2nd_union": ("bar", {"some_field": 4}),
                }
            ],
        ),
        (
            [
                {
                    "my_1st_union": {"some_field": 5, "-type": "bar"},
                    "my_2nd_union": {"some_field": 6, "-type": "foo"},
                }
            ],
            [
                {
                    "my_1st_union": ("bar", {"some_field": 5}),
                    "my_2nd_union": ("foo", {"some_field": 6}),
                }
            ],
        ),
        (
            [
                {
                    "my_1st_union": {"some_field": 7, "-type": "bar"},
                    "my_2nd_union": {"some_field": 8, "-type": "bar"},
                }
            ],
            [
                {
                    "my_1st_union": ("bar", {"some_field": 7}),
                    "my_2nd_union": ("bar", {"some_field": 8}),
                }
            ],
        ),
    ],
)
def test_union_path_picked_with_record_type_hint(input_records, expected_roundtrip):
    """https://github.com/fastavro/fastavro/issues/540"""
    schema = {
        "type": "record",
        "name": "my_record",
        "fields": [
            {
                "name": "my_1st_union",
                "type": [
                    {
                        "name": "foo",
                        "type": "record",
                        "fields": [{"name": "some_field", "type": "int"}],
                    },
                    {
                        "name": "bar",
                        "type": "record",
                        "fields": [{"name": "some_field", "type": "int"}],
                    },
                ],
            },
            {"name": "my_2nd_union", "type": ["foo", "bar"]},
        ],
    }

    rt_records = roundtrip(
        fastavro.parse_schema(schema), input_records, return_record_name=True
    )
    assert expected_roundtrip == rt_records


def test_non_string_types_raise_type_error():
    """https://github.com/fastavro/fastavro/issues/556"""

    schema = {
        "type": "record",
        "fields": [{"name": "field", "type": "string"}],
        "name": "test_non_string_types_raise_type_error",
    }

    record = {"field": None}

    new_file = BytesIO()
    with pytest.raises(TypeError, match="must be string"):
        fastavro.schemaless_writer(new_file, schema, record)


def test_name_changes_to_full_record():
    """https://github.com/fastavro/fastavro/issues/601"""
    writer_schema = {
        "type": "record",
        "name": "test",
        "fields": [
            {
                "name": "test1",
                "type": {
                    "type": "record",
                    "name": "my_record",
                    "fields": [{"name": "field1", "type": "string"}],
                },
            },
            {"name": "test2", "type": "my_record"},
        ],
    }

    reader_schema = {
        "type": "record",
        "name": "test",
        "fields": [
            {
                "name": "test2",
                "type": {
                    "type": "record",
                    "name": "my_record",
                    "fields": [{"name": "field1", "type": "string"}],
                },
            },
        ],
    }

    records = [{"test1": {"field1": "foo"}, "test2": {"field1": "bar"}}]
    output = roundtrip(writer_schema, records, reader_schema=reader_schema)
    assert output == [{"test2": {"field1": "bar"}}]


def test_record_names_must_match():
    """https://github.com/fastavro/fastavro/issues/601"""
    writer_schema = {
        "type": "record",
        "name": "test",
        "fields": [
            {
                "name": "test1",
                "type": {
                    "type": "record",
                    "name": "my_record",
                    "fields": [{"name": "field1", "type": "string"}],
                },
            },
            {"name": "test2", "type": "my_record"},
        ],
    }

    reader_schema = {
        "type": "record",
        "name": "test",
        "fields": [
            {
                "name": "test2",
                "type": {
                    "type": "record",
                    "name": "different_name",
                    "fields": [{"name": "field1", "type": "string"}],
                },
            },
        ],
    }

    records = [{"test1": {"field1": "foo"}, "test2": {"field1": "bar"}}]
    with pytest.raises(SchemaResolutionError):
        roundtrip(writer_schema, records, reader_schema=reader_schema)


def test_strict_option():
    """https://github.com/fastavro/fastavro/issues/549"""
    schema = {
        "namespace": "namespace",
        "name": "name",
        "type": "record",
        "fields": [
            {"name": "field_1", "type": "boolean"},
            {"name": "field_2", "type": ["null", "string"], "default": None},
        ],
    }

    test_record1 = {"field_1": True, "field_2": "foo", "field_3": "something"}
    test_record2 = {"field_1": True}
    test_record3 = {"field_2": "foo"}

    with pytest.raises(ValueError, match="field_3"):
        roundtrip(schema, [test_record1], writer_kwargs={"strict": True})

    with pytest.raises(ValueError, match="field_2 is specified .*? but missing"):
        roundtrip(schema, [test_record2], writer_kwargs={"strict": True})

    with pytest.raises(ValueError, match="field_1 is specified .*? but missing"):
        roundtrip(schema, [test_record3], writer_kwargs={"strict": True})


def test_strict_allow_default_option():
    """https://github.com/fastavro/fastavro/issues/549"""
    schema = {
        "namespace": "namespace",
        "name": "name",
        "type": "record",
        "fields": [
            {"name": "field_1", "type": "boolean"},
            {"name": "field_2", "type": ["null", "string"], "default": None},
        ],
    }

    test_record1 = {"field_1": True, "field_2": "foo", "field_3": "something"}
    test_record2 = {"field_1": True}
    test_record3 = {"field_2": "foo"}

    with pytest.raises(ValueError, match="field_3"):
        roundtrip(schema, [test_record1], writer_kwargs={"strict_allow_default": True})

    roundtrip(schema, [test_record2], writer_kwargs={"strict_allow_default": True})

    with pytest.raises(ValueError, match="field_1 is specified .*? but missing"):
        roundtrip(schema, [test_record3], writer_kwargs={"strict_allow_default": True})


def test_disable_tuple_notation_option():
    """https://github.com/fastavro/fastavro/issues/548"""
    schema = {
        "namespace": "namespace",
        "name": "name",
        "type": "record",
        "fields": [
            {"name": "foo", "type": ["string", {"type": "array", "items": "string"}]}
        ],
    }

    new_records = roundtrip(
        schema,
        [{"foo": ("string", "0")}],
        writer_kwargs={"disable_tuple_notation": True},
    )
    assert new_records == [{"foo": ["string", "0"]}]


def test_strict_allow_default_bug():
    """https://github.com/fastavro/fastavro/issues/638"""
    schema = {
        "namespace": "namespace",
        "name": "name",
        "type": "record",
        "fields": [{"name": "some_field", "type": "string", "default": "test"}],
    }

    test_record = {"eggs": "eggs"}

    with pytest.raises(ValueError, match="record contains more fields .*? eggs"):
        roundtrip(schema, [test_record], writer_kwargs={"strict_allow_default": True})


def test_nan_default_value():
    """https://github.com/fastavro/fastavro/issues/674"""
    schema = {
        "namespace": "namespace",
        "name": "name",
        "type": "record",
        "fields": [{"name": "some_field", "type": "float", "default": "nan"}],
    }

    test_record = {}

    result_value = roundtrip(schema, [test_record])[0]["some_field"]
    assert math.isnan(result_value)


def test_allow_union_with_bad_default_if_correct_reader_schema_present():
    """https://github.com/fastavro/fastavro/issues/676"""
    writer_schema = {
        "type": "record",
        "name": "Test",
        "namespace": "test",
        "fields": [{"name": "union", "type": ["null", "int"], "default": 0}],
    }

    reader_schema = {
        "type": "record",
        "name": "Test",
        "namespace": "test",
        "fields": [{"name": "union", "type": ["int", "null"], "default": 0}],
    }

    records = [{}]

    # Parse the schema and ignore union default errors like it would have been
    # parsed in the past
    parsed_writer_schema = fastavro.parse_schema(
        writer_schema, _ignore_default_error=True
    )

    roundtrip_records = roundtrip(
        parsed_writer_schema, records, reader_schema=reader_schema
    )
    assert roundtrip_records == [{"union": 0}]


def test_allow_bad_default_if_correct_reader_schema_present():
    """https://github.com/fastavro/fastavro/issues/676"""
    writer_schema = {
        "type": "record",
        "name": "Test",
        "namespace": "test",
        "fields": [{"name": "field1", "type": "int", "default": None}],
    }

    reader_schema = {
        "type": "record",
        "name": "Test",
        "namespace": "test",
        "fields": [{"name": "field1", "type": "int", "default": 0}],
    }

    records = [{"field1": 5}]

    # Parse the schema and ignore union default errors like it would have been
    # parsed in the past
    parsed_writer_schema = fastavro.parse_schema(
        writer_schema, _ignore_default_error=True
    )

    roundtrip_records = roundtrip(
        parsed_writer_schema, records, reader_schema=reader_schema
    )
    assert roundtrip_records == [{"field1": 5}]



from fastavro import writer as fastavro_writer
from fastavro.read import SchemaResolutionError
from fastavro.utils import generate_one
import fastavro

import pytest

from io import BytesIO

schema_dict_a = {
    "namespace": "example.avro2",
    "type": "record",
    "name": "evtest",
    "fields": [{"name": "a", "type": "int"}],
}

record_a = {"a": 123}

schema_dict_a_b = {
    "namespace": "example.avro2",
    "type": "record",
    "name": "evtest",
    "fields": [
        {"name": "a", "type": "int"},
        {"name": "b", "type": ["null", "int"], "default": None},
    ],
}

record_a_b = {"a": 234, "b": 345}

schema_dict_a_c = {
    "namespace": "example.avro2",
    "type": "record",
    "name": "evtest",
    "fields": [{"name": "a", "type": "int"}, {"name": "c", "type": ["null", "int"]}],
}


def avro_to_bytes_with_schema(avro_schema, avro_dict):
    with BytesIO() as bytes_io:
        fastavro_writer(bytes_io, avro_schema, [avro_dict])
        return bytes_io.getvalue()


def bytes_with_schema_to_avro(avro_read_schema, binary):
    with BytesIO(binary) as bytes_io:
        reader = fastavro.reader(bytes_io, avro_read_schema)
        return next(reader)


def test_evolution_drop_field():
    record_bytes_a_b = avro_to_bytes_with_schema(schema_dict_a_b, record_a_b)
    record_a = bytes_with_schema_to_avro(schema_dict_a, record_bytes_a_b)
    assert "b" not in record_a


def test_evolution_add_field_with_default():
    record_bytes_a = avro_to_bytes_with_schema(schema_dict_a, record_a)
    record_b = bytes_with_schema_to_avro(schema_dict_a_b, record_bytes_a)
    assert "b" in record_b
    assert record_b.get("b") is None


def test_evolution_add_field_without_default():
    with pytest.raises(SchemaResolutionError):
        record_bytes_a = avro_to_bytes_with_schema(schema_dict_a, record_a)
        bytes_with_schema_to_avro(schema_dict_a_c, record_bytes_a)


def test_enum_evolution_no_default_failure():
    original_schema = {
        "type": "enum",
        "name": "test",
        "symbols": ["FOO", "BAR"],
    }

    new_schema = {
        "type": "enum",
        "name": "test",
        "symbols": ["BAZ", "BAR"],
    }

    original_records = ["FOO"]

    bio = BytesIO()
    fastavro.writer(bio, original_schema, original_records)
    bio.seek(0)

    with pytest.raises(fastavro.read.SchemaResolutionError):
        list(fastavro.reader(bio, new_schema))


def test_enum_evolution_using_default():
    original_schema = {
        "type": "enum",
        "name": "test",
        "symbols": ["A", "B"],
    }

    new_schema = {
        "type": "enum",
        "name": "test",
        "symbols": ["C", "D"],
        "default": "C",
    }

    original_records = ["A"]

    bio = BytesIO()
    fastavro.writer(bio, original_schema, original_records)
    bio.seek(0)

    new_records = list(fastavro.reader(bio, new_schema))
    assert new_records == ["C"]


def test_schema_matching_with_records_in_arrays():
    """https://github.com/fastavro/fastavro/issues/363"""
    original_schema = {
        "type": "record",
        "name": "DataRecord",
        "fields": [
            {
                "name": "string1",
                "type": "string",
            },
            {
                "name": "subrecord",
                "type": {
                    "type": "array",
                    "items": {
                        "type": "record",
                        "name": "SubRecord",
                        "fields": [
                            {
                                "name": "string2",
                                "type": "string",
                            }
                        ],
                    },
                },
            },
        ],
    }

    new_schema = {
        "type": "record",
        "name": "DataRecord",
        "fields": [
            {
                "name": "string1",
                "type": "string",
            },
            {
                "name": "subrecord",
                "type": {
                    "type": "array",
                    "items": {
                        "type": "record",
                        "name": "SubRecord",
                        "fields": [
                            {
                                "name": "string2",
                                "type": "string",
                            },
                            {
                                "name": "logs",
                                "default": None,
                                "type": [
                                    "null",
                                    {
                                        "type": "array",
                                        "items": {
                                            "type": "record",
                                            "name": "LogRecord",
                                            "fields": [
                                                {
                                                    "name": "msg",
                                                    "type": "string",
                                                    "default": "",
                                                }
                                            ],
                                        },
                                    },
                                ],
                            },
                        ],
                    },
                },
            },
        ],
    }

    record = {
        "string1": "test",
        "subrecord": [{"string2": "foo"}],
    }

    binary = avro_to_bytes_with_schema(original_schema, record)

    output_using_original_schema = bytes_with_schema_to_avro(original_schema, binary)
    assert output_using_original_schema == record

    output_using_new_schema = bytes_with_schema_to_avro(new_schema, binary)
    assert output_using_new_schema == {
        "string1": "test",
        "subrecord": [{"string2": "foo", "logs": None}],
    }


def test_schema_migrate_record_to_union():
    """https://github.com/fastavro/fastavro/issues/406"""
    original_schema = {
        "name": "Item",
        "type": "record",
        "fields": [
            {
                "name": "category",
                "type": {
                    "type": "record",
                    "name": "Category",
                    "fields": [{"name": "name", "type": "string"}],
                },
            }
        ],
    }

    new_schema_record_first = {
        "name": "Item",
        "type": "record",
        "fields": [
            {
                "name": "category",
                "type": [
                    {
                        "type": "record",
                        "name": "Category",
                        "fields": [{"name": "name", "type": "string"}],
                    },
                    "null",
                ],
            }
        ],
    }

    new_schema_null_first = {
        "name": "Item",
        "type": "record",
        "fields": [
            {
                "name": "category",
                "type": [
                    "null",
                    {
                        "type": "record",
                        "name": "Category",
                        "fields": [{"name": "name", "type": "string"}],
                    },
                ],
            }
        ],
    }

    record = {"category": {"name": "my-category"}}

    binary = avro_to_bytes_with_schema(original_schema, record)

    output_using_original_schema = bytes_with_schema_to_avro(original_schema, binary)
    assert output_using_original_schema == record

    output_using_new_schema_record_first = bytes_with_schema_to_avro(
        new_schema_record_first, binary
    )
    assert output_using_new_schema_record_first == record

    output_using_new_schema_null_first = bytes_with_schema_to_avro(
        new_schema_null_first, binary
    )
    assert output_using_new_schema_null_first == record


def test_union_of_lists_evolution_with_doc():
    """https://github.com/fastavro/fastavro/issues/486"""
    original_schema = {
        "name": "test_union_of_lists_evolution_with_doc",
        "type": "record",
        "fields": [
            {
                "name": "id",
                "type": [
                    "null",
                    {
                        "name": "some_record",
                        "type": "record",
                        "fields": [{"name": "field", "type": "string"}],
                    },
                ],
            }
        ],
    }

    new_schema = {
        "name": "test_union_of_lists_evolution_with_doc",
        "type": "record",
        "fields": [
            {
                "name": "id",
                "type": [
                    "null",
                    {
                        "name": "some_record",
                        "type": "record",
                        "doc": "some documentation",
                        "fields": [{"name": "field", "type": "string"}],
                    },
                ],
            }
        ],
    }

    record = {"id": {"field": "foo"}}

    binary = avro_to_bytes_with_schema(original_schema, record)

    output_using_new_schema = bytes_with_schema_to_avro(new_schema, binary)
    assert output_using_new_schema == record


def test_union_of_lists_evolution_with_extra_type():
    """https://github.com/fastavro/fastavro/issues/486"""
    original_schema = {
        "name": "test_union_of_lists_evolution_with_extra_type",
        "type": "record",
        "fields": [
            {
                "name": "id",
                "type": [
                    "null",
                    {
                        "name": "some_record",
                        "type": "record",
                        "fields": [{"name": "field", "type": "string"}],
                    },
                ],
            }
        ],
    }

    new_schema = {
        "name": "test_union_of_lists_evolution_with_extra_type",
        "type": "record",
        "fields": [
            {
                "name": "id",
                "type": [
                    "null",
                    {
                        "name": "some_record",
                        "type": "record",
                        "fields": [{"name": "field", "type": "string"}],
                    },
                    "string",
                ],
            }
        ],
    }

    record = {"id": {"field": "foo"}}

    binary = avro_to_bytes_with_schema(original_schema, record)

    output_using_new_schema = bytes_with_schema_to_avro(new_schema, binary)
    assert output_using_new_schema == record


INT_ARRAY = {"type": "array", "items": "int"}
LONG_ARRAY = {"type": "array", "items": "long"}
INT_MAP = {"type": "map", "values": "int"}
LONG_MAP = {"type": "map", "values": "long"}
ENUM_AB = {"type": "enum", "name": "enum", "symbols": ["A", "B"]}
ENUM2_AB = {"type": "enum", "name": "enum2", "symbols": ["A", "B"]}
ENUM_ABC = {"type": "enum", "name": "enum", "symbols": ["A", "B", "C"]}
INT_UNION = ["int"]
LONG_UNION = ["long"]
FLOAT_UNION = ["float"]
DOUBLE_UNION = ["double"]
STRING_UNION = ["string"]
BYTES_UNION = ["bytes"]
INT_LONG_UNION = ["int", "long"]
INT_FLOAT_UNION = ["int", "float"]
INT_LONG_FLOAT_DOUBLE_UNION = ["int", "long", "float", "double"]
INT_STRING_UNION = ["int", "string"]
STRING_INT_UNION = ["string", "int"]
FIXED_4_BYTES = {"type": "fixed", "name": "fixed", "size": 4}
FIXED_8_BYTES = {"type": "fixed", "name": "fixed", "size": 8}
EMPTY_RECORD = {"type": "record", "name": "record", "fields": []}
A_INT_RECORD = {
    "type": "record",
    "name": "record",
    "fields": [{"name": "a", "type": "int"}],
}
A_DINT_RECORD = {
    "type": "record",
    "name": "record",
    "fields": [{"name": "a", "type": "int", "default": 0}],
}
A_LONG_RECORD = {
    "type": "record",
    "name": "record",
    "fields": [{"name": "a", "type": "long"}],
}
A_INT_B_INT_RECORD = {
    "type": "record",
    "name": "record",
    "fields": [{"name": "a", "type": "int"}, {"name": "b", "type": "int"}],
}
A_INT_B_DINT_RECORD = {
    "type": "record",
    "name": "record",
    "fields": [
        {"name": "a", "type": "int"},
        {"name": "b", "type": "int", "default": 0},
    ],
}
A_DINT_B_DINT_RECORD = {
    "type": "record",
    "name": "record",
    "fields": [
        {"name": "a", "type": "int", "default": 0},
        {"name": "b", "type": "int", "default": 0},
    ],
}
ENUM_ABC_ENUM_DEFAULT_A_RECORD = {
    "type": "record",
    "name": "record",
    "fields": [
        {
            "name": "Field",
            "type": {
                "type": "enum",
                "name": "enum",
                "symbols": ["A", "B", "C"],
                "default": "A",
            },
        },
    ],
}
ENUM_AB_ENUM_DEFAULT_A_RECORD = {
    "type": "record",
    "name": "record",
    "fields": [
        {
            "name": "Field",
            "type": {
                "type": "enum",
                "name": "enum",
                "symbols": ["A", "B"],
                "default": "A",
            },
        },
    ],
}
ENUM_ABC_FIELD_DEFAULT_B_ENUM_DEFAULT_A_RECORD = {
    "type": "record",
    "name": "record",
    "fields": [
        {
            "name": "Field",
            "type": {
                "type": "enum",
                "name": "enum",
                "symbols": ["A", "B", "C"],
                "default": "A",
            },
            "default": "B",
        },
    ],
}
ENUM_AB_FIELD_DEFAULT_A_ENUM_DEFAULT_B_RECORD = {
    "type": "record",
    "name": "record",
    "fields": [
        {
            "name": "Field",
            "type": {
                "type": "enum",
                "name": "enum",
                "symbols": ["A", "B"],
                "default": "B",
            },
            "default": "A",
        }
    ],
}
NS_RECORD1 = {
    "type": "record",
    "name": "record",
    "fields": [
        {
            "name": "f1",
            "type": [
                "null",
                {
                    "type": "array",
                    "items": {
                        "type": "record",
                        "name": "inner_record",
                        "namespace": "ns1",
                        "fields": [{"name": "a", "type": "int"}],
                    },
                },
            ],
        }
    ],
}
NS_RECORD2 = {
    "type": "record",
    "name": "record",
    "fields": [
        {
            "name": "f1",
            "type": [
                "null",
                {
                    "type": "array",
                    "items": {
                        "type": "record",
                        "name": "inner_record",
                        "namespace": "ns2",
                        "fields": [{"name": "a", "type": "int"}],
                    },
                },
            ],
        }
    ],
}


@pytest.mark.parametrize(
    "writer_schema,reader_schema",
    [
        ("boolean", "boolean"),
        ("int", "int"),
        ("int", "long"),
        ("long", "long"),
        ("int", "float"),
        ("long", "float"),
        ("int", "double"),
        ("long", "double"),
        ("float", "double"),
        ("string", "string"),
        ("bytes", "bytes"),
        ("string", "bytes"),
        # ("bytes", "string"),
        (INT_ARRAY, INT_ARRAY),
        (INT_ARRAY, LONG_ARRAY),
        (INT_MAP, INT_MAP),
        (INT_MAP, LONG_MAP),
        (ENUM_AB, ENUM_AB),
        (ENUM_AB, ENUM_ABC),
        (INT_UNION, FLOAT_UNION),
        (LONG_UNION, FLOAT_UNION),
        (INT_LONG_UNION, FLOAT_UNION),
        (INT_UNION, INT_UNION),
        (STRING_INT_UNION, INT_STRING_UNION),
        (INT_UNION, LONG_UNION),
        (INT_UNION, FLOAT_UNION),
        (INT_UNION, DOUBLE_UNION),
        (LONG_UNION, FLOAT_UNION),
        (LONG_UNION, DOUBLE_UNION),
        (FLOAT_UNION, DOUBLE_UNION),
        (STRING_UNION, BYTES_UNION),
        # (BYTES_UNION, STRING_UNION),
        (INT_FLOAT_UNION, DOUBLE_UNION),
        (INT_FLOAT_UNION, "float"),
        (INT_LONG_UNION, "long"),
        (INT_FLOAT_UNION, "double"),
        (INT_LONG_FLOAT_DOUBLE_UNION, "double"),
        (FIXED_4_BYTES, FIXED_4_BYTES),
        (EMPTY_RECORD, EMPTY_RECORD),
        (A_INT_RECORD, EMPTY_RECORD),
        (A_INT_RECORD, A_INT_RECORD),
        (A_INT_RECORD, A_DINT_RECORD),
        (A_DINT_RECORD, A_DINT_RECORD),
        (A_DINT_RECORD, A_INT_RECORD),
        (A_INT_RECORD, A_LONG_RECORD),
        (A_INT_B_INT_RECORD, A_INT_RECORD),
        (A_INT_B_INT_RECORD, A_DINT_RECORD),
        (A_INT_RECORD, A_INT_B_DINT_RECORD),
        (EMPTY_RECORD, A_DINT_B_DINT_RECORD),
        (A_INT_RECORD, A_DINT_B_DINT_RECORD),
        (A_DINT_B_DINT_RECORD, A_INT_B_INT_RECORD),
        ({"type": "null"}, {"type": "null"}),
        ("null", "null"),
        (ENUM_ABC_ENUM_DEFAULT_A_RECORD, ENUM_AB_ENUM_DEFAULT_A_RECORD),
        (
            ENUM_ABC_FIELD_DEFAULT_B_ENUM_DEFAULT_A_RECORD,
            ENUM_AB_FIELD_DEFAULT_A_ENUM_DEFAULT_B_RECORD,
        ),
        (A_DINT_B_DINT_RECORD, A_INT_B_INT_RECORD),
        (NS_RECORD2, NS_RECORD1),
    ],
)
def test_schema_compatibility(writer_schema, reader_schema):
    bio = BytesIO()
    writer_data = generate_one(writer_schema)
    fastavro.writer(bio, writer_schema, [writer_data])
    bio.seek(0)

    # This should not throw an exception if the schemas are compatible
    list(fastavro.reader(bio, reader_schema))


def test_bytes_writer_string_reader():
    writer_schema = "bytes"
    reader_schema = "string"

    # If the bytes are utf-8 encoded, reader should work
    bio = BytesIO()
    fastavro.writer(bio, writer_schema, [b"123"])
    bio.seek(0)
    list(fastavro.reader(bio, reader_schema))

    # If the bytes are not utf-8, decoding will fail
    bio = BytesIO()
    fastavro.writer(bio, writer_schema, ["かわいい".encode("cp932")])
    bio.seek(0)
    with pytest.raises(UnicodeDecodeError):
        list(fastavro.reader(bio, reader_schema))


@pytest.mark.parametrize(
    "writer_schema,reader_schema",
    [
        ("int", "null"),
        ("long", "null"),
        ("int", "boolean"),
        ("null", "int"),
        ("boolean", "int"),
        ("long", "int"),
        ("float", "int"),
        ("double", "int"),
        ("float", "long"),
        ("double", "long"),
        ("double", "float"),
        ("string", "double"),
        ("string", FIXED_4_BYTES),
        ("boolean", "string"),
        ("int", "string"),
        ("null", "bytes"),
        ("int", "bytes"),
        ("int", A_INT_RECORD),
        (LONG_ARRAY, INT_ARRAY),
        (INT_ARRAY, INT_MAP),
        (INT_MAP, INT_ARRAY),
        (LONG_MAP, INT_MAP),
        (ENUM2_AB, "int"),
        ("int", ENUM2_AB),
        (FIXED_4_BYTES, FIXED_8_BYTES),
        (FIXED_8_BYTES, FIXED_4_BYTES),
    ],
)
def test_schema_incompatibility(writer_schema, reader_schema):
    bio = BytesIO()
    writer_data = generate_one(writer_schema)
    fastavro.writer(bio, writer_schema, [writer_data])
    bio.seek(0)

    with pytest.raises(SchemaResolutionError):
        list(fastavro.reader(bio, reader_schema))


def test_union_writer_simple_reader():
    writer_schema = INT_LONG_FLOAT_DOUBLE_UNION
    writer_data = ("float", generate_one("float"))
    for reader_schema in INT_LONG_FLOAT_DOUBLE_UNION:
        bio = BytesIO()
        fastavro.writer(bio, writer_schema, [writer_data])
        bio.seek(0)

        # If we are reading as an int or long, it should fail. If we are reading
        # as a float or double, it should pass
        if reader_schema in ("int", "long"):
            with pytest.raises(SchemaResolutionError):
                list(fastavro.reader(bio, reader_schema))
        else:
            list(fastavro.reader(bio, reader_schema))


def test_writer_enum_more_symbols_than_reader_enum():
    writer_schema = ENUM_ABC
    reader_schema = ENUM_AB

    for symbol in writer_schema["symbols"]:
        bio = BytesIO()
        fastavro.writer(bio, writer_schema, [symbol])
        bio.seek(0)

        # If the symbol is in the reader symbols, it should work. Otherwise it
        # should fail
        if symbol in reader_schema["symbols"]:
            list(fastavro.reader(bio, reader_schema))
        else:
            with pytest.raises(SchemaResolutionError):
                list(fastavro.reader(bio, reader_schema))


def test_union_of_schemas_evolution():
    """https://github.com/fastavro/fastavro/issues/553"""
    original_schema = [
        {
            "name": "other_record",
            "type": "record",
            "fields": [
                {"name": "other_one", "type": "boolean"},
                {"name": "other_two", "type": "float"},
            ],
        },
        {
            "name": "root_record",
            "type": "record",
            "fields": [
                {"name": "one", "type": "boolean"},
                {"name": "two", "type": "float"},
                {"name": "three", "type": "other_record"},
            ],
        },
    ]

    new_schema = [
        {
            "name": "other_record",
            "type": "record",
            "fields": [
                {"name": "other_two", "type": "float"},
                {"name": "new_three", "type": "float", "default": 0.0},
            ],
        },
        {
            "name": "root_record",
            "type": "record",
            "fields": [{"name": "three", "type": "other_record"}],
        },
    ]

    record = {"one": True, "two": 0.0, "three": {"other_one": False, "other_two": 1.0}}

    binary = avro_to_bytes_with_schema(original_schema, record)

    output_using_new_schema = bytes_with_schema_to_avro(new_schema, binary)
    assert output_using_new_schema == {"three": {"other_two": 1.0, "new_three": 0.0}}


def test_records_match_by_unnamespaced_name():
    """https://issues.apache.org/jira/browse/AVRO-3561"""
    original_schema = {
        "name": "ns.test_record",
        "type": "record",
        "fields": [{"name": "f1", "type": "int"}],
    }

    new_schema = {
        "name": "ns.foo.test_record",
        "type": "record",
        "fields": [
            {"name": "f1", "type": "int"},
            {"name": "f2", "type": "int", "default": 3},
        ],
    }

    record = {"f1": 0}

    binary = avro_to_bytes_with_schema(original_schema, record)

    output_using_new_schema = bytes_with_schema_to_avro(new_schema, binary)
    assert output_using_new_schema == {"f1": 0, "f2": 3}



import random
from io import BytesIO
from fastavro import schemaless_writer
from fastavro.utils import generate_one, generate_many, anonymize_schema


def test_generate():
    schema = {
        "type": "record",
        "name": "Test",
        "namespace": "test",
        "fields": [
            {"name": "null", "type": "null"},
            {"name": "boolean", "type": "boolean"},
            {"name": "string", "type": "string"},
            {"name": "bytes", "type": "bytes"},
            {"name": "int", "type": "int"},
            {"name": "long", "type": "long"},
            {"name": "float", "type": "float"},
            {"name": "double", "type": "double"},
            {
                "name": "fixed",
                "type": {"type": "fixed", "name": "fixed_field", "size": 5},
            },
            {
                "name": "union",
                "type": [
                    "null",
                    "int",
                    {
                        "type": "record",
                        "name": "union_record",
                        "fields": [{"name": "union_record_field", "type": "string"}],
                    },
                ],
            },
            {
                "name": "enum",
                "type": {
                    "type": "enum",
                    "name": "enum_field",
                    "symbols": ["FOO", "BAR"],
                },
            },
            {"name": "array", "type": {"type": "array", "items": "string"}},
            {"name": "map", "type": {"type": "map", "values": "int"}},
            {
                "name": "record",
                "type": {
                    "type": "record",
                    "name": "subrecord",
                    "fields": [{"name": "sub_int", "type": "int"}],
                },
            },
            {"name": "named_type", "type": "subrecord"},
        ],
    }

    count = 10

    # Use list() to exhaust the generator)
    assert len(list(generate_many(schema, count))) == count


def test_anonymize():
    schema = {
        "type": "record",
        "name": "Test",
        "namespace": "test",
        "doc": "this is a record",
        "fields": [
            {"name": "null", "type": "null"},
            {"name": "boolean", "type": "boolean"},
            {"name": "string", "type": "string", "default": "foo"},
            {"name": "bytes", "type": "bytes", "aliases": ["alias_field"]},
            {"name": "int", "type": "int", "doc": "doc"},
            {"name": "long", "type": "long"},
            {"name": "float", "type": "float"},
            {"name": "double", "type": "double"},
            {
                "name": "fixed",
                "type": {"type": "fixed", "name": "fixed_field", "size": 5},
            },
            {
                "name": "union",
                "type": [
                    "null",
                    "int",
                    {
                        "type": "record",
                        "name": "union_record",
                        "fields": [{"name": "union_record_field", "type": "string"}],
                    },
                ],
            },
            {"name": "array", "type": {"type": "array", "items": "string"}},
            {"name": "map", "type": {"type": "map", "values": "int"}},
            {
                "name": "record",
                "type": {
                    "type": "record",
                    "name": "subrecord",
                    "fields": [{"name": "sub_int", "type": "int"}],
                },
            },
            {"name": "named_type", "type": "subrecord"},
            {"name": "other_int", "type": {"type": "int"}},
        ],
    }

    anonymous_schema = anonymize_schema(schema)

    # Maintain random state so that other tests continue to be based off the
    # main starting seed
    seed_state = random.getstate()
    random.seed(1)
    record = generate_one(schema)

    random.seed(1)
    anonymous_record = generate_one(anonymous_schema)
    random.setstate(seed_state)

    bio1 = BytesIO()
    schemaless_writer(bio1, schema, record)

    bio2 = BytesIO()
    schemaless_writer(bio2, anonymous_schema, anonymous_record)

    assert bio1.getvalue() == bio2.getvalue()


def test_enum_symbols_get_anonymized():
    schema = {
        "type": "enum",
        "name": "enum_field",
        "symbols": ["FOO", "BAR"],
    }

    anonymous_schema = anonymize_schema(schema)

    assert anonymous_schema["symbols"] != schema["symbols"]



import datetime
from io import BytesIO

import fastavro
from fastavro.const import MCS_PER_SECOND, MLS_PER_SECOND
from fastavro._logical_writers_py import prepare_timestamp_micros
from fastavro._logical_writers_py import prepare_timestamp_millis
import pytest
from .conftest import assert_naive_datetime_equal_to_tz_datetime
from datetime import timezone


schema = {
    "fields": [
        {
            "name": "timestamp-millis",
            "type": {"type": "long", "logicalType": "timestamp-millis"},
        },
        {
            "name": "timestamp-micros",
            "type": {"type": "long", "logicalType": "timestamp-micros"},
        },
    ],
    "namespace": "namespace",
    "name": "name",
    "type": "record",
}


# Test Time Zone with fixed offset and no DST
class TSTTzinfo(datetime.tzinfo):
    def utcoffset(self, dt):
        return datetime.timedelta(hours=10)

    def tzname(self, dt):
        return "TST"

    def dst(self, dt):
        return datetime.timedelta(0)


tst = TSTTzinfo()


def serialize(schema, data):
    bytes_writer = BytesIO()
    fastavro.schemaless_writer(bytes_writer, schema, data)
    return bytes_writer.getvalue()


def deserialize(schema, binary):
    bytes_writer = BytesIO()
    bytes_writer.write(binary)
    bytes_writer.seek(0)

    res = fastavro.schemaless_reader(bytes_writer, schema)
    return res


def timestamp_mls_from_timestamp(timestamp):
    # need to remove precision from microseconds field
    mls = int(timestamp.microsecond / 1000) * 1000
    return datetime.datetime(
        year=timestamp.year,
        month=timestamp.month,
        day=timestamp.day,
        hour=timestamp.hour,
        minute=timestamp.minute,
        second=timestamp.second,
        microsecond=mls,
        tzinfo=timestamp.tzinfo,
    )


def test_tz_attributes():
    assert tst.tzname(None) == "TST"
    assert tst.utcoffset(None) != timezone.utc.utcoffset(None)


@pytest.fixture(scope="session")
def timestamp_data():
    timestamp = datetime.datetime.now(tz=timezone.utc)
    return {
        "timestamp-millis": timestamp_mls_from_timestamp(timestamp),
        "timestamp-micros": timestamp,
    }


@pytest.fixture(scope="session")
def timestamp_data_naive():
    timestamp = datetime.datetime.now()
    return {
        "timestamp-millis": timestamp_mls_from_timestamp(timestamp),
        "timestamp-micros": timestamp,
    }


@pytest.fixture(scope="session")
def read_data(timestamp_data):
    binary = serialize(schema, timestamp_data)
    return deserialize(schema, binary)


@pytest.fixture(scope="session")
def read_data_naive(timestamp_data_naive):
    binary = serialize(schema, timestamp_data_naive)
    return deserialize(schema, binary)


def test_timestamp_micros_tz_input(timestamp_data, read_data):
    original = timestamp_data["timestamp-micros"]
    assert original.tzinfo is not None
    read = read_data["timestamp-micros"]
    assert read.tzinfo is not None
    assert original == read
    read_in_test_tz = read.astimezone(tst)
    assert original == read_in_test_tz


def test_timestamp_millis_tz_input(timestamp_data, read_data):
    original = timestamp_data["timestamp-millis"]
    assert original.tzinfo is not None
    read = read_data["timestamp-millis"]
    assert read.tzinfo is not None
    assert original == read
    read_in_test_tz = read.astimezone(tst)
    assert original == read_in_test_tz


def test_timestamp_micros_naive_input(timestamp_data_naive, read_data_naive):
    original = timestamp_data_naive["timestamp-micros"]
    assert original.tzinfo is None
    read = read_data_naive["timestamp-micros"]
    assert read.tzinfo is not None
    assert_naive_datetime_equal_to_tz_datetime(original, read)


def test_timestamp_millis_naive_input(timestamp_data_naive, read_data_naive):
    original = timestamp_data_naive["timestamp-millis"]
    assert original.tzinfo is None
    read = read_data_naive["timestamp-millis"]
    assert read.tzinfo is not None
    assert_naive_datetime_equal_to_tz_datetime(original, read)


def test_prepare_timestamp_micros():
    # seconds from epoch == 1234567890
    reference_time = datetime.datetime(2009, 2, 13, 23, 31, 30, tzinfo=timezone.utc)
    mcs_from_epoch = 1234567890 * MCS_PER_SECOND
    assert prepare_timestamp_micros(reference_time, schema) == mcs_from_epoch
    timestamp_tst = reference_time.astimezone(tst)
    assert prepare_timestamp_micros(reference_time, schema) == prepare_timestamp_micros(
        timestamp_tst, schema
    )


def test_prepare_timestamp_millis():
    # seconds from epoch == 1234567890
    reference_time = datetime.datetime(2009, 2, 13, 23, 31, 30, tzinfo=timezone.utc)
    mcs_from_epoch = 1234567890 * MLS_PER_SECOND
    assert prepare_timestamp_millis(reference_time, schema) == mcs_from_epoch
    timestamp_tst = reference_time.astimezone(tst)
    assert prepare_timestamp_millis(reference_time, schema) == prepare_timestamp_millis(
        timestamp_tst, schema
    )


@pytest.mark.parametrize(
    "my_date",
    [
        datetime.datetime(1974, 4, 4, 0, 0, 0, 1000, tzinfo=timezone.utc),
        datetime.datetime(2515, 1, 1, 0, 0, 0, 37000, tzinfo=timezone.utc),
        datetime.datetime(881, 1, 1, 0, 0, 0, 257000, tzinfo=timezone.utc),
        datetime.datetime(2243, 1, 1, 0, 0, 0, 64000, tzinfo=timezone.utc),
    ],
)
def test_problematic_timestamp_millis(my_date):
    schema = {
        "fields": [
            {
                "name": "timestamp-millis",
                "type": {"type": "long", "logicalType": "timestamp-millis"},
            },
            {
                "name": "timestamp-micros",
                "type": {"type": "long", "logicalType": "timestamp-micros"},
            },
        ],
        "name": "test_problematic_timestamp_millis",
        "type": "record",
    }

    binary = serialize(
        schema, {"timestamp-millis": my_date, "timestamp-micros": my_date}
    )
    roundtrip_data = deserialize(schema, binary)

    assert my_date == roundtrip_data["timestamp-millis"]
    assert my_date == roundtrip_data["timestamp-micros"]


@pytest.mark.parametrize(
    "my_date",
    [
        datetime.datetime(1974, 4, 4, 0, 0, 0, 1000),
        datetime.datetime(2515, 1, 1, 0, 0, 0, 37000),
        datetime.datetime(2243, 1, 1, 0, 0, 0, 64000),
    ],
)
def test_problematic_timestamp_millis_naive_time(my_date):
    schema = {
        "fields": [
            {
                "name": "timestamp-millis",
                "type": {"type": "long", "logicalType": "timestamp-millis"},
            },
            {
                "name": "timestamp-micros",
                "type": {"type": "long", "logicalType": "timestamp-micros"},
            },
        ],
        "name": "test_problematic_timestamp_millis",
        "type": "record",
    }

    binary = serialize(
        schema, {"timestamp-millis": my_date, "timestamp-micros": my_date}
    )
    roundtrip_data = deserialize(schema, binary)

    assert_naive_datetime_equal_to_tz_datetime(
        my_date, roundtrip_data["timestamp-millis"]
    )
    assert_naive_datetime_equal_to_tz_datetime(
        my_date, roundtrip_data["timestamp-micros"]
    )



import fastavro

import pytest

from io import BytesIO


schema = {
    "fields": [
        {"name": "str_null", "type": ["null", "string"]},
        {"name": "str", "type": "string"},
        {"name": "integ_null", "type": ["null", "int"]},
        {"name": "integ", "type": "int"},
    ],
    "namespace": "namespace",
    "name": "missingerror",
    "type": "record",
}


def serialize(schema, *records):
    buffer = BytesIO()
    fastavro.writer(buffer, schema, records)
    serialized = buffer.getvalue()
    return serialized


def test_types_match():
    records = [
        {
            "str_null": "str",
            "str": "str",
            "integ_null": 21,
            "integ": 21,
        }
    ]
    serialize(schema, *records)


def test_string_in_int_raises():
    records = [
        {
            "str_null": "str",
            "str": "str",
            "integ_null": "str",
            "integ": 21,
        }
    ]

    with pytest.raises(ValueError):
        serialize(schema, *records)


def test_string_in_int_null_raises():
    records = [
        {
            "str_null": "str",
            "str": "str",
            "integ_null": 11,
            "integ": "str",
        }
    ]
    with pytest.raises(TypeError):
        serialize(schema, *records)


def test_int_in_string_null_raises():
    records = [
        {
            "str_null": 11,
            "str": "str",
            "integ_null": 21,
            "integ": 21,
        }
    ]
    with pytest.raises(ValueError):
        serialize(schema, *records)


def test_int_in_string_raises():
    records = [
        {
            "str_null": "str",
            "str": 11,
            "integ_null": 21,
            "integ": 21,
        }
    ]

    # Raises AttributeError on py2 and TypeError on py3
    with pytest.raises((TypeError, AttributeError)):
        serialize(schema, *records)


@pytest.mark.parametrize(  #
    ("value", "binary"),  #
    [
        (0, b"\x00"),
        (-1, b"\x01"),
        (1, b"\x02"),
        (-2, b"\x03"),
        (2, b"\x04"),
        (2147483647, b"\xfe\xff\xff\xff\x0f"),
        (-2147483648, b"\xff\xff\xff\xff\x0f"),
        (9223372036854775807, b"\xfe\xff\xff\xff\xff\xff\xff\xff\xff\x01"),
        (-9223372036854775808, b"\xff\xff\xff\xff\xff\xff\xff\xff\xff\x01"),
    ],
)
def test_int_binary(value, binary):
    schema = {"type": "long"}
    buffer = BytesIO()

    fastavro.schemaless_writer(buffer, schema, value)
    assert buffer.getvalue() == binary, "Invalid integer encoding."

    deserialized = fastavro.schemaless_reader(BytesIO(binary), schema)
    assert deserialized == value, "Invalid integer decoding."



# cython: language_level=3
# cython: auto_cpdef=True

import hashlib
from io import StringIO
from os import path
from copy import deepcopy
import json
from libc.math cimport floor, log10
import re

from .repository import FlatDictRepository, SchemaRepositoryError
from .const import AVRO_TYPES
from ._schema_common import (
    PRIMITIVES,
    UnknownType,
    SchemaParseException,
    RESERVED_PROPERTIES,
    OPTIONAL_FIELD_PROPERTIES,
    RESERVED_FIELD_PROPERTIES,
    JAVA_FINGERPRINT_MAPPING,
    FINGERPRINT_ALGORITHMS,
    RABIN_64,
    rabin_fingerprint,
)

SYMBOL_REGEX = re.compile(r"[A-Za-z_][A-Za-z0-9_]*")
NO_DEFAULT = object()


cpdef inline is_nullable_union(schema):
    count = 0
    for s in schema:
        extracted_type = extract_record_type(s)
        if extracted_type not in AVRO_TYPES or extracted_type == "record":
            count += 1

    return count == 1


cpdef inline extract_record_type(schema):
    if isinstance(schema, dict):
        return schema["type"]

    if isinstance(schema, list):
        return "union"

    return schema  


cpdef inline str extract_logical_type(schema):
    if not isinstance(schema, dict):
        return None
    rt = schema["type"]
    lt = schema.get("logicalType")
    if lt:
        # TODO: Building this string every time is going to be relatively slow.
        return f"{rt}-{lt}"
    return None


cpdef fullname(schema):
    return schema_name(schema, "")[1]


cpdef schema_name(schema, parent_ns):
    try:
        name = schema["name"]
    except KeyError:
        raise SchemaParseException(
            f'"name" is a required field missing from the schema: {schema}'
        )

    namespace = schema.get("namespace", parent_ns)
    if not namespace:
        return namespace, name
    elif "." in name:
        return "", name
    else:
        return namespace, f"{namespace}.{name}"


cpdef expand_schema(schema):
    return parse_schema(schema, expand=True, _write_hint=False)


def parse_schema(
    schema,
    named_schemas=None,
    *,
    expand=False,
    _write_hint=True,
    _force=False,
    _ignore_default_error=False,
):
    if named_schemas is None:
        named_schemas = {}

    if isinstance(schema, dict) and "__fastavro_parsed" in schema:
        if "__named_schemas" in schema:
            for key, value in schema["__named_schemas"].items():
                named_schemas[key] = value
        else:
            # Some old schemas might only have __fastavro_parsed and not
            # __named_schemas since that came later. For these schemas, we need
            # to re-parse the schema to handle named types
            return _parse_schema(
                schema,
                "",
                expand,
                _write_hint,
                set(),
                named_schemas,
                NO_DEFAULT,
                _ignore_default_error,
            )

    if _force or expand:
        return _parse_schema(
            schema,
            "",
            expand,
            _write_hint,
            set(),
            named_schemas,
            NO_DEFAULT,
            _ignore_default_error,
        )
    elif isinstance(schema, dict) and "__fastavro_parsed" in schema:
        return schema
    elif isinstance(schema, list):
        # If we are given a list we should make sure that the immediate sub
        # schemas have the hint in them
        return [
            parse_schema(
                s,
                named_schemas,
                expand=expand,
                _write_hint=_write_hint,
                _force=_force,
                _ignore_default_error=_ignore_default_error,
            )
            for s in schema
        ]
    else:
        return _parse_schema(
            schema,
            "",
            expand,
            _write_hint,
            set(),
            named_schemas,
            NO_DEFAULT,
            _ignore_default_error,
        )


cdef _raise_default_value_error(
    default, schema_type, in_union, ignore_default_error
):
    if ignore_default_error:
        return
    elif in_union:
        text = f"first schema in union with type: {schema_type}"
    else:
        text = f"schema type: {schema_type}"

    raise SchemaParseException(f"Default value <{default}> must match {text}")


cdef _parse_schema(
    schema,
    namespace,
    expand,
    _write_hint,
    names,
    named_schemas,
    default,
    ignore_default_error,
    in_union=False,
):
    # union schemas
    if isinstance(schema, list):
        parsed_schemas = []
        for index, s in enumerate(schema):
            if index == 0:
                parsed_schemas.append(
                    _parse_schema(
                        s,
                        namespace,
                        expand,
                        False,
                        names,
                        named_schemas,
                        default,
                        ignore_default_error,
                        in_union=True,
                    )
                )
            else:
                parsed_schemas.append(
                    _parse_schema(
                        s,
                        namespace,
                        expand,
                        False,
                        names,
                        named_schemas,
                        NO_DEFAULT,
                        ignore_default_error,
                    )
                )
        return parsed_schemas

    # string schemas; this could be either a named schema or a primitive type
    elif not isinstance(schema, dict):
        if schema in PRIMITIVES:
            if default is not NO_DEFAULT:
                if (
                    (schema == "null" and default is not None)
                    or (schema == "boolean" and not isinstance(default, bool))
                    or (schema == "string" and not isinstance(default, str))
                    or (schema == "bytes" and not isinstance(default, str))
                    or (schema == "double" and not isinstance(float(default), float))
                    or (schema == "float" and not isinstance(float(default), float))
                    or (schema == "int" and not isinstance(default, int))
                    or (schema == "long" and not isinstance(default, int))
                ):
                    _raise_default_value_error(
                        default, schema, in_union, ignore_default_error
                    )
            return schema

        if "." not in schema and namespace:
            schema = namespace + "." + schema

        if schema not in named_schemas:
            raise UnknownType(schema)
        elif expand:
            # If `name` is in the schema, it has been fully resolved and so we
            # can include the full schema. If `name` is not in the schema yet,
            # then we are still recursing that schema and must use the named
            # schema or else we will have infinite recursion when printing the
            # final schema
            if "name" in named_schemas[schema]:
                return named_schemas[schema]
            else:
                return schema
        else:
            return schema

    else:
        # Remaining valid schemas must be dict types
        schema_type = schema["type"]

        parsed_schema = {
            key: value
            for key, value in schema.items()
            if key not in RESERVED_PROPERTIES
        }
        parsed_schema["type"] = schema_type

        if "doc" in schema:
            parsed_schema["doc"] = schema["doc"]

        # Correctness checks for logical types
        logical_type = parsed_schema.get("logicalType")
        if logical_type == "decimal":
            scale = parsed_schema.get("scale")
            if scale and (not isinstance(scale, int) or scale < 0):
                raise SchemaParseException(
                    f"decimal scale must be a positive integer, not {scale}"
                )

            precision = parsed_schema.get("precision")
            if precision:
                if not isinstance(precision, int) or precision <= 0:
                    raise SchemaParseException(
                        "decimal precision must be a positive integer, "
                        + f"not {precision}"
                    )
                if schema_type == "fixed":
                    # https://avro.apache.org/docs/current/spec.html#Decimal
                    size = schema["size"]
                    max_precision = int(floor(log10(2) * (8 * size - 1)))
                    if precision > max_precision:
                        raise SchemaParseException(
                            f"decimal precision of {precision} doesn't fit "
                            + f"into array of length {size}"
                        )

            if scale and precision and precision < scale:
                raise SchemaParseException(
                    "decimal scale must be less than or equal to "
                    + f"the precision of {precision}"
                )

        if schema_type == "array":
            parsed_schema["items"] = _parse_schema(
                schema["items"],
                namespace,
                expand,
                False,
                names,
                named_schemas,
                NO_DEFAULT,
                ignore_default_error,
            )
            if default is not NO_DEFAULT and not isinstance(default, list):
                _raise_default_value_error(
                    default, schema_type, in_union, ignore_default_error
                )

        elif schema_type == "map":
            parsed_schema["values"] = _parse_schema(
                schema["values"],
                namespace,
                expand,
                False,
                names,
                named_schemas,
                NO_DEFAULT,
                ignore_default_error,
            )
            if default is not NO_DEFAULT and not isinstance(default, dict):
                _raise_default_value_error(
                    default, schema_type, in_union, ignore_default_error
                )

        elif schema_type == "enum":
            _, fullname = schema_name(schema, namespace)
            if fullname in names:
                raise SchemaParseException(f"redefined named type: {fullname}")
            names.add(fullname)

            _validate_enum_symbols(schema)

            if default is not NO_DEFAULT and not isinstance(default, str):
                _raise_default_value_error(
                    default, schema_type, in_union, ignore_default_error
                )

            named_schemas[fullname] = parsed_schema

            parsed_schema["name"] = fullname
            parsed_schema["symbols"] = schema["symbols"]

        elif schema_type == "fixed":
            _, fullname = schema_name(schema, namespace)
            if fullname in names:
                raise SchemaParseException(f"redefined named type: {fullname}")
            names.add(fullname)

            if default is not NO_DEFAULT and not isinstance(default, str):
                _raise_default_value_error(
                    default, schema_type, in_union, ignore_default_error
                )

            named_schemas[fullname] = parsed_schema

            parsed_schema["name"] = fullname
            parsed_schema["size"] = schema["size"]

        elif schema_type == "record" or schema_type == "error":
            # records
            namespace, fullname = schema_name(schema, namespace)
            if fullname in names:
                raise SchemaParseException(f"redefined named type: {fullname}")
            names.add(fullname)

            if default is not NO_DEFAULT and not isinstance(default, dict):
                _raise_default_value_error(
                    default, schema_type, in_union, ignore_default_error
                )

            named_schemas[fullname] = parsed_schema

            fields = []
            for field in schema.get("fields", []):
                fields.append(
                    parse_field(
                        field,
                        namespace,
                        expand,
                        names,
                        named_schemas,
                        ignore_default_error,
                    )
                )

            parsed_schema["name"] = fullname
            parsed_schema["fields"] = fields

            # Hint that we have parsed the record
            if _write_hint:
                # Make a copy of parsed_schema so that we don't have a cyclical
                # reference. Using deepcopy is pretty slow, and we don't need a
                # true deepcopy so this works good enough
                named_schemas[fullname] = {
                    k: v for k, v in parsed_schema.items()
                }

                parsed_schema["__fastavro_parsed"] = True
                parsed_schema["__named_schemas"] = named_schemas

        elif schema_type in PRIMITIVES:
            parsed_schema["type"] = schema_type
            if default is not NO_DEFAULT:
                if (
                    (schema_type == "null" and default is not None)
                    or (schema_type == "boolean" and not isinstance(default, bool))
                    or (schema_type == "string" and not isinstance(default, str))
                    or (schema_type == "bytes" and not isinstance(default, str))
                    or (schema_type == "double" and not isinstance(default, float))
                    or (schema_type == "float" and not isinstance(default, float))
                    or (schema_type == "int" and not isinstance(default, int))
                    or (schema_type == "long" and not isinstance(default, int))
                ):
                    _raise_default_value_error(
                        default, schema_type, in_union, ignore_default_error
                    )

        else:
            raise UnknownType(schema)

        return parsed_schema


cdef parse_field(field, namespace, expand, names, named_schemas, ignore_default_error):
    parsed_field = {
        key: value
        for key, value in field.items()
        if key not in RESERVED_FIELD_PROPERTIES
    }

    for prop in OPTIONAL_FIELD_PROPERTIES:
        if prop in field:
            parsed_field[prop] = field[prop]

    # Aliases must be a list
    aliases = parsed_field.get("aliases", [])
    if not isinstance(aliases, list):
        msg = f"aliases must be a list, not {aliases}"
        raise SchemaParseException(msg)

    default = field.get("default", NO_DEFAULT)

    parsed_field["name"] = field["name"]
    parsed_field["type"] = _parse_schema(
        field["type"],
        namespace,
        expand,
        False,
        names,
        named_schemas,
        default,
        ignore_default_error,
    )

    return parsed_field


def load_schema(
    schema_path,
    *,
    repo=None,
    named_schemas=None,
    _write_hint=True,
    _injected_schemas=None,
):
    schema_name = schema_path
    if repo is None:
        file_dir, file_name = path.split(schema_path)
        schema_name, _file_ext = path.splitext(file_name)
        repo = FlatDictRepository(file_dir)

    if named_schemas is None:
        named_schemas = {}

    if _injected_schemas is None:
        _injected_schemas = set()

    return _load_schema(
        schema_name, repo, named_schemas, _write_hint, _injected_schemas
    )


cdef _load_schema(schema_name, repo, named_schemas, write_hint, injected_schemas):
    try:
        schema = repo.load(schema_name)
        return _parse_schema_with_repo(
            schema,
            repo,
            named_schemas,
            write_hint,
            injected_schemas,
        )
    except SchemaRepositoryError as error:
        raise error


cdef _parse_schema_with_repo(
    schema,
    repo,
    named_schemas,
    write_hint,
    injected_schemas,
):
    try:
        schema_copy = deepcopy(named_schemas)
        return parse_schema(schema, named_schemas=named_schemas, _write_hint=write_hint)
    except UnknownType as error:
        missing_subject = error.name
        try:
            sub_schema = _load_schema(
                missing_subject,
                repo,
                named_schemas=schema_copy,
                write_hint=False,
                injected_schemas=injected_schemas,
            )
        except SchemaRepositoryError:
            raise error

        if sub_schema["name"] not in injected_schemas:
            injected_schema = _inject_schema(schema, sub_schema)
            if isinstance(schema, str) or isinstance(schema, list):
                schema = injected_schema[0]
            injected_schemas.add(sub_schema["name"])
        return _parse_schema_with_repo(
            schema, repo, schema_copy, write_hint, injected_schemas
        )


cdef _inject_schema(outer_schema, inner_schema, namespace="", is_injected=False):
    # Once injected, we can stop checking to see if we need to inject since it
    # should only be done once at most
    if is_injected is True:
        return outer_schema, is_injected

    # union schemas
    if isinstance(outer_schema, list):
        union = []
        for each_schema in outer_schema:
            if is_injected:
                union.append(each_schema)
            else:
                return_schema, injected = _inject_schema(
                    each_schema, inner_schema, namespace, is_injected
                )
                union.append(return_schema)
                if injected is True:
                    is_injected = injected
        return union, is_injected

    # string schemas; this could be either a named schema or a primitive type
    elif not isinstance(outer_schema, dict):
        if outer_schema in PRIMITIVES:
            return outer_schema, is_injected

        if "." not in outer_schema and namespace:
            outer_schema = namespace + "." + outer_schema

        if outer_schema == inner_schema["name"]:
            return inner_schema, True
        else:
            # Hit a named schema that has already been loaded previously. Return
            # the outer_schema so we keep looking
            return outer_schema, is_injected
    else:
        # Remaining valid schemas must be dict types
        schema_type = outer_schema["type"]

        if schema_type == "array":
            return_schema, injected = _inject_schema(
                outer_schema["items"], inner_schema, namespace, is_injected
            )
            outer_schema["items"] = return_schema
            return outer_schema, injected

        elif schema_type == "map":
            return_schema, injected = _inject_schema(
                outer_schema["values"], inner_schema, namespace, is_injected
            )
            outer_schema["values"] = return_schema
            return outer_schema, injected

        elif schema_type == "enum":
            return outer_schema, is_injected

        elif schema_type == "fixed":
            return outer_schema, is_injected

        elif schema_type == "record" or schema_type == "error":
            # records
            namespace, _ = schema_name(outer_schema, namespace)
            fields = []
            for field in outer_schema.get("fields", []):
                if is_injected:
                    fields.append(field)
                else:
                    return_schema, injected = _inject_schema(
                        field["type"], inner_schema, namespace, is_injected
                    )
                    field["type"] = return_schema
                    fields.append(field)

                    if injected is True:
                        is_injected = injected
            if fields:
                outer_schema["fields"] = fields

            return outer_schema, is_injected

        elif schema_type in PRIMITIVES:
            return outer_schema, is_injected

        else:
            raise Exception(
                "Internal error; "
                + "You should raise an issue in the fastavro github repository"
            )


def load_schema_ordered(ordered_schemas, *, _write_hint=True):
    loaded_schemas = []
    named_schemas = {}
    for idx, schema_path in enumerate(ordered_schemas):
        # _write_hint is always False except maybe the outer most schema
        _last = _write_hint if idx + 1 == len(ordered_schemas) else False
        schema = load_schema(
            schema_path, named_schemas=named_schemas, _write_hint=_last
        )
        loaded_schemas.append(schema)

    top_first_order = loaded_schemas[::-1]
    outer_schema = top_first_order.pop(0)

    while top_first_order:
        sub_schema = top_first_order.pop(0)
        _inject_schema(outer_schema, sub_schema)

    return outer_schema


def to_parsing_canonical_form(schema):
    fo = StringIO()
    _to_parsing_canonical_form(parse_schema(schema), fo)
    return fo.getvalue()


cdef _to_parsing_canonical_form(schema, fo):
    # union schemas
    if isinstance(schema, list):
        fo.write("[")
        for idx, s in enumerate(schema):
            if idx != 0:
                fo.write(",")
            _to_parsing_canonical_form(s, fo)
        fo.write("]")

    # string schemas; this could be either a named schema or a primitive type
    elif not isinstance(schema, dict):
        fo.write(f'"{schema}"')

    else:
        # Remaining valid schemas must be dict types
        schema_type = schema["type"]

        if schema_type == "array":
            fo.write(f'{{"type":"{schema_type}","items":')
            _to_parsing_canonical_form(schema["items"], fo)
            fo.write("}")

        elif schema_type == "map":
            fo.write(f'{{"type":"{schema_type}","values":')
            _to_parsing_canonical_form(schema["values"], fo)
            fo.write("}")

        elif schema_type == "enum":
            name = schema["name"]
            fo.write(f'{{"name":"{name}","type":"{schema_type}","symbols":[')

            for idx, symbol in enumerate(schema["symbols"]):
                if idx != 0:
                    fo.write(",")
                fo.write(f'"{symbol}"')
            fo.write("]}")

        elif schema_type == "fixed":
            name = schema["name"]
            size = schema["size"]
            fo.write(f'{{"name":"{name}","type":"{schema_type}","size":{size}}}')

        elif schema_type == "record" or schema_type == "error":
            name = schema["name"]
            fo.write(f'{{"name":"{name}","type":"record","fields":[')

            for idx, field in enumerate(schema["fields"]):
                if idx != 0:
                    fo.write(",")
                name = field["name"]
                fo.write(f'{{"name":"{name}","type":')
                _to_parsing_canonical_form(field["type"], fo)
                fo.write("}")
            fo.write("]}")

        elif schema_type in PRIMITIVES:
            fo.write(f'"{schema_type}"')


def fingerprint(parsing_canonical_form, algorithm):
    if algorithm not in FINGERPRINT_ALGORITHMS:
        raise ValueError(
            f"Unknown schema fingerprint algorithm {algorithm}. "
            + f"Valid values include: {FINGERPRINT_ALGORITHMS}"
        )

    # Fix Java names
    algorithm = JAVA_FINGERPRINT_MAPPING.get(algorithm, algorithm)

    if algorithm == RABIN_64:
        return rabin_fingerprint(parsing_canonical_form.encode())

    h = hashlib.new(algorithm, parsing_canonical_form.encode())
    return h.hexdigest()


def _validate_enum_symbols(schema):
    symbols = schema["symbols"]
    for symbol in symbols:
        if not isinstance(symbol, str) or not re.match(SYMBOL_REGEX, symbol):
            raise SchemaParseException(
                "Every symbol must match the regular expression [A-Za-z_][A-Za-z0-9_]*"
            )
    if len(symbols) != len(set(symbols)):
        raise SchemaParseException("All symbols in an enum must be unique")

    if "default" in schema and schema["default"] not in symbols:
        raise SchemaParseException("Default value for enum must be in symbols list")



from io import BytesIO
import fastavro

import pytest


def roundtrip(schema, records, new_schema):
    new_file = BytesIO()
    fastavro.writer(new_file, schema, records)
    new_file.seek(0)

    reader = fastavro.reader(new_file, new_schema)
    new_records = list(reader)
    return new_records


def test_aliases_not_present():
    schema = {
        "type": "record",
        "name": "test_aliases_not_present",
        "fields": [{"name": "test", "type": "double"}],
    }

    new_schema = {
        "type": "record",
        "name": "test_aliases_not_present",
        "fields": [
            {"name": "newtest", "type": "double", "aliases": ["testX"]},
        ],
    }

    records = [{"test": 1.2}]

    with pytest.raises(fastavro.read.SchemaResolutionError):
        roundtrip(schema, records, new_schema)


def test_incompatible_aliases():
    schema = {
        "type": "record",
        "name": "test_incompatible_aliases",
        "fields": [{"name": "test", "type": "double"}],
    }

    new_schema = {
        "type": "record",
        "name": "test_incompatible_aliases",
        "fields": [
            {"name": "newtest", "type": "int", "aliases": ["test"]},
        ],
    }

    records = [{"test": 1.2}]

    with pytest.raises(fastavro.read.SchemaResolutionError):
        roundtrip(schema, records, new_schema)


def test_aliases_in_reader_schema():
    schema = {
        "type": "record",
        "name": "test_aliases_in_reader_schema",
        "fields": [{"name": "test", "type": "int"}],
    }

    new_schema = {
        "type": "record",
        "name": "test_aliases_in_reader_schema",
        "fields": [{"name": "newtest", "type": "int", "aliases": ["test"]}],
    }

    records = [{"test": 1}]

    assert roundtrip(schema, records, new_schema) == [{"newtest": 1}]


def test_aliases_with_default_value_and_field_added():
    """https://github.com/fastavro/fastavro/issues/225"""
    schema = {
        "type": "record",
        "name": "test_aliases_with_default_value",
        "fields": [{"name": "test", "type": "int"}],
    }

    new_schema = {
        "type": "record",
        "name": "test_aliases_with_default_value",
        "fields": [
            {"name": "newtest", "type": "int", "default": 0, "aliases": ["test"]},
            {"name": "test2", "type": "int", "default": 100},
        ],
    }

    records = [{"test": 1}]

    new_records = roundtrip(schema, records, new_schema)
    assert new_records == [{"newtest": 1, "test2": 100}]


def test_record_name_alias():
    schema = {
        "type": "record",
        "name": "test_record_name_alias",
        "fields": [{"name": "test", "type": "int"}],
    }

    new_schema = {
        "type": "record",
        "name": "test_record_name_alias_new",
        "aliases": "test_record_name_alias",
        "fields": [{"name": "test", "type": "int"}],
    }

    records = [{"test": 1}]

    assert roundtrip(schema, records, new_schema) == [{"test": 1}]


def test_fixed_name_alias():
    schema = {"type": "fixed", "name": "test_fixed_name_alias", "size": 4}

    new_schema = {
        "type": "fixed",
        "name": "test_fixed_name_alias_new",
        "aliases": "test_fixed_name_alias",
        "size": 4,
    }

    records = [b"1234"]

    assert roundtrip(schema, records, new_schema) == [b"1234"]


def test_enum_name_alias():
    schema = {"type": "enum", "name": "test_enum_name_alias", "symbols": ["FOO"]}

    new_schema = {
        "type": "enum",
        "name": "test_enum_name_alias_new",
        "aliases": "test_enum_name_alias",
        "symbols": ["FOO"],
    }

    records = ["FOO"]

    assert roundtrip(schema, records, new_schema) == ["FOO"]
